import logging
import threading
import time
import warnings
from html import escape
from io import BytesIO

import matplotlib
import numpy as np
import pylab as p
from pyramid.response import Response
from pyramid.view import view_config

from ... import core
from ...config import (webview_cache_time, webview_default_image_format,
                       webview_plots_dpi)
from ...log import logger
from ...util.cache_dict import CacheDict
from . import halo_from_request, simulation_from_request, timestep_from_request

matplotlib.use('agg')

_matplotlib_lock = threading.RLock()

def sessionfree_lru_cache(num):
    """A drop-in replacement for functools.lru_cache that works with sqlalchemy ORM objects as first argument

    The purpose is to cache results of retrieving/calculating data from a database. However, such calculations
    rely on having an ORM object, which itself may change between calls because a different session is in use
    (e.g. if different threads are involved). This decorator roughly mimics the behaviour of functools.lru_cache,
    but keys on the .id attribute of the first argument, rather than the argument itself.
    """
    cache = CacheDict(cache_len=num)
    def f(func):
        def g(*args):
            key = type(args[0]),args[0].id, *args[1:]
            if key in cache:
                return cache[key]
            else:
                v = func(*args)
                cache[key] = v
                return v

        return g
    return f

def decode_property_name(name):
    name = name.replace("_slash_","/")
    return name

def format_array(data, max_array_length=3):
    if len(data)>max_array_length:
        return "Array"
    data_fmt = []
    for d in data:
        data_fmt.append(format_data(d))
    return "["+(", ".join(data_fmt))+"]"

def format_number(data):
    if np.issubdtype(type(data), np.integer):
        return "%d" % data
    elif np.issubdtype(type(data), np.floating):
        if abs(data) > 1e5 or abs(data) < 1e-2:
            return "%.2e" % data
        else:
            return "%.2f" % data
    else:
        warnings.warn("Formatting unknown number type")
        return str(data)

def format_data(data, request=None, relative_to=None, max_array_length=3):
    try:
        if hasattr(data,'__len__'):
            return format_array(data, max_array_length)
        elif np.issubdtype(type(data), np.number):
            return format_number(data)
        elif isinstance(data, core.SimulationObjectBase):
            return format_halo(data, request, relative_to)
        else:
            return escape(repr(data))
    except Exception as e:
        logging.exception("Exception in format_data")
        return str(e)



def _relative_description(this_halo, other_halo) :
    if other_halo is None :
        return "null"
    elif this_halo and this_halo.id==other_halo.id:
        return "this"
    elif this_halo and this_halo.timestep_id == other_halo.timestep_id :
        return "%s %d"%(other_halo.tag,other_halo.halo_number)
    elif this_halo and this_halo.timestep.simulation_id == other_halo.timestep.simulation_id :
        return "%s %d at t=%.2e Gyr"%(other_halo.tag,other_halo.halo_number, other_halo.timestep.time_gyr)
    else :
        if (not this_halo) or abs(this_halo.timestep.time_gyr - other_halo.timestep.time_gyr)>0.001:
            return "%s %d in %8s at t=%.2e Gyr"%(other_halo.tag,other_halo.halo_number, other_halo.timestep.simulation.basename,
                                                   other_halo.timestep.time_gyr)
        else:
            return "%s %d in %8s"%(other_halo.tag,other_halo.halo_number, other_halo.timestep.simulation.basename)


def format_halo(halo, request, relative_to=None):
    if relative_to==halo or request is None:
        return _relative_description(relative_to, halo)
    else:
        link = request.route_url('halo_view', simid=halo.timestep.simulation.escaped_basename,
                                 timestepid=halo.timestep.escaped_extension,
                                 halonumber=halo.basename)
        return "<a href='%s'>%s</a>"%(link, _relative_description(relative_to, halo))

def is_number(data):
    return np.issubdtype(type(data), np.number)

def can_use_elements_in_plot(data_array):
    if len(data_array)==0:
        return False
    else:
        return is_number(data_array[0])

def is_boolean(data):
    return np.issubdtype(type(data), np.bool_) and not np.issubdtype(type(data), np.number) and not hasattr(data,'__len__')

def can_use_elements_as_filter(data_array):
    if len(data_array)==0:
        return False
    else:
        return is_boolean(data_array[0])

def is_array(data):
    return isinstance(data, np.ndarray) and data.ndim>0

def elements_are_arrays(data_array):
    if len(data_array)==0:
        return False
    else:
        return is_array(data_array[0])


@view_config(route_name='calculate_all', renderer='json', http_cache=webview_cache_time)
def calculate_all(request):
    ts = timestep_from_request(request)
    typetag = request.matchdict['typetag']
    try:
        data, = ts.calculate_all(decode_property_name(request.matchdict['nameid']),
                                 sanitize=False, order_by_halo_number=True, object_type=typetag)
    except Exception as e:
        logging.exception("Exception in calculate_all")
        return {'error': getattr(e,'message',""), 'error_class': type(e).__name__}

    return {'timestep': ts.escaped_extension, 'data_formatted': [format_data(d, request) for d in data],
            'is_number': can_use_elements_in_plot(data),
            'is_boolean': can_use_elements_as_filter(data),
            'is_array': elements_are_arrays(data)}

@view_config(route_name='get_property', renderer='json', http_cache=webview_cache_time)
def get_property(request):
    halo = halo_from_request(request)
    property_name = decode_property_name(request.matchdict['nameid'])

    return get_property_data(halo, property_name, request)


def get_property_data(halo, property_name_or_result, request=None):
    if isinstance(property_name_or_result, str):
        try:
            result, p_info = _get_property_from_halo_and_name(halo, property_name_or_result)
            name = property_name_or_result
        except Exception as e:
            logging.exception("Exception in get_property_data")
            return {'error': getattr(e, 'message', ""), 'error_class': type(e).__name__}
    else:
        result = property_name_or_result.data_raw
        name = property_name_or_result.name.text

    return {'data_formatted': format_data(result, request, halo),
            'mini_language_query': name,
            'is_number': is_number(result),
            'is_boolean': is_boolean(result),
            'is_array': is_array(result)}


def start(request):
    p.ioff()
    p.figure(figsize=(float(request.GET.get('width',1000))/webview_plots_dpi,
                   float(request.GET.get('height',1000))/webview_plots_dpi))
    request.canvas = p.get_current_fig_manager().canvas


CONTENT_TYPES = {
    'png': 'image/png',
    'svg': 'image/svg+xml',
    'pdf': 'application/pdf'
}

def finish(request, getImage=True):
    extension = request.matchdict.get("ext", webview_default_image_format)
    print(f'Got extension: {extension}')
    if getImage:
        enter_finish_time = time.time()
        request.canvas.draw()
        draw_time = time.time()
        buffer = BytesIO()
        p.savefig(buffer, format=extension, dpi=webview_plots_dpi, bbox_inches='tight')
        end_time = time.time()

        logger.info(
            "Image rendering: matplotlib %.2fs; %s conversion %.3fs",
            draw_time - enter_finish_time,
            extension.upper(),
            end_time - draw_time
        )

    p.close()

    if getImage:
        try:
            r = Response(
                content_type=CONTENT_TYPES[extension],
                body=buffer.getvalue()
            )
            r.cache_expires(webview_cache_time)
            return r
        except KeyError:
            raise NotImplementedError(
                'Tangos does not support the provided image format: '
                f'{ext}. '
                'This can be changed in the config.'
            )


def rescale_plot(request):
    logx = request.GET.get('logx',False)
    logy = request.GET.get('logy',False)
    if logx and logy:
        p.loglog()
    elif logx:
        p.semilogx()
    elif logy:
        p.semilogy()

@view_config(route_name='gathered_plot')
def gathered_plot(request):
    start_time = time.time()
    name1, name2, v1, v2 = gathered_plot_data_from_request(request)
    logger.info("Gathering data took %.2fs"%(time.time()-start_time))
    with _matplotlib_lock:
        start(request)
        p.plot(v1,v2,'k.')
        p.xlabel(name1)
        p.ylabel(name2)
        rescale_plot(request)
        return finish(request)


@view_config(route_name='gathered_csv', renderer='csv')
def gathered_csv(request):
    name1, name2, v1, v2 = gathered_plot_data_from_request(request)
    return {
        'header': [name1, name2],
        'rows': np.array((v1,v2)).T,
        'name': "timestep_" + name1 + "_vs_" + name2
    }

def gathered_plot_data_from_request(request):
    ts = timestep_from_request(request)
    name1 = decode_property_name(request.matchdict['nameid1'])
    name2 = decode_property_name(request.matchdict['nameid2'])
    filter = decode_property_name(request.GET.get('filter', ""))
    object_typetag = request.GET.get('object_typetag', None)

    v1, v2 = _gathered_plot_data_from_parameters(ts, name1, name2, filter, object_typetag)

    return name1, name2, v1, v2

@sessionfree_lru_cache(100)
def _gathered_plot_data_from_parameters(ts, name1, name2, filter, object_typetag):
    if filter != "":
        v1, v2, f = ts.calculate_all(name1, name2, filter, object_typetag=object_typetag)
        v1 = v1[f]
        v2 = v2[f]
    else:
        v1, v2 = ts.calculate_all(name1, name2, object_typetag=object_typetag)
    return v1, v2


@view_config(route_name='cascade_plot')
def cascade_plot(request):
    name1, name2, v1, v2 = cascade_plot_data_from_request(request)
    with _matplotlib_lock:
        start(request)
        p.plot(v1,v2,'k')
        p.xlabel(name1)
        p.ylabel(name2)
        rescale_plot(request)
        return finish(request)

@view_config(route_name='cascade_csv', renderer='csv')
def cascade_csv(request):
    name1, name2, v1, v2 = cascade_plot_data_from_request(request)
    return {
        'header': [name1, name2],
        'rows': np.array((v1,v2)).T,
        'name': "timeseries_"+name1+"_vs_"+name2
    }

def cascade_plot_data_from_request(request):
    halo = halo_from_request(request)
    name1 = decode_property_name(request.matchdict['nameid1'])
    name2 = decode_property_name(request.matchdict['nameid2'])
    v1, v2 = _cascade_plot_data_from_parameters(halo, name1, name2)
    return name1, name2, v1, v2

@sessionfree_lru_cache(100)
def _cascade_plot_data_from_parameters(halo, name1, name2):
    v1, v2 = halo.calculate_for_progenitors(name1, name2)
    return v1, v2


def _sanitize_lims(val, absolute, data, default, log):
    if val is None:
        val = default
    val = float(val)

    if absolute:
        return val
    elif log:
        return np.nanpercentile(data[data>0], val)
    else:
        return np.nanpercentile(data, val)


def image_plot(request, data, property_info):
    log = request.GET.get('logimage', "0") == "1"
    vmin, vmax = (request.GET.get(_, None) for _ in ("vmin", "vmax"))
    cmap = request.GET.get("cmap", None)
    absolute = request.GET.get("absolute", "0") == "1"
    with _matplotlib_lock:
        start(request)

        if property_info:
            width = property_info.plot_extent()
        else:
            width = 1.0

        # This is required to properly use log norms with ranges larger than 10^7
        data = data.astype(np.float64)

        if data.ndim == 2:
            vmin = _sanitize_lims(vmin, absolute, data, 0, log)
            vmax = _sanitize_lims(vmax, absolute, data, 100, log)

            vmin, vmax = min(vmin, vmax), max(vmin, vmax)

        kwa = {"cmap": cmap}

        if log:
            kwa["norm"] = p.matplotlib.colors.LogNorm(vmin, vmax)
        else:
            kwa["norm"] = p.matplotlib.colors.Normalize(vmin, vmax)

        if width is not None:
            if hasattr(width, '__len__'):
                kwa["extent"] = width
            else:
                kwa["extent"] = (-width/2, width/2, -width/2, width/2)

        p.imshow(data, **kwa)

        if property_info:
            add_xy_labels(property_info, request)

        if data.ndim == 2:
            cb = p.colorbar()
            if property_info and property_info.plot_clabel() :
                cb.set_label(property_info.plot_clabel())

        return finish(request)


def add_xy_labels(property_info, request):
    p.xlabel(property_info.plot_xlabel())
    ylabel = property_info.plot_ylabel()
    # cludge follows - should be eliminated by fixing the mess around multi-name vs single-name property classes
    if not isinstance(ylabel, str):
        try:
            ylabel = ylabel[property_info.index_of_name(decode_property_name(request.matchdict['nameid']))]
        except:
            ylabel = ""
    p.ylabel(ylabel)


@view_config(route_name='array_plot')
def array_plot(request):
    halo = halo_from_request(request)
    name = decode_property_name(request.matchdict['nameid'])

    val, property_info = _get_property_from_halo_and_name(halo, name)

    if len(val.shape)>1:
        return image_plot(request, val, property_info)

    with _matplotlib_lock:
        start(request)

        p.plot(property_info.plot_x_values(val),val)

        if property_info.plot_xlog() and property_info.plot_ylog():
            p.loglog()
        elif property_info.plot_xlog():
            p.semilogx()
        elif property_info.plot_ylog():
            p.semilogy()


        if property_info.plot_yrange():
            p.ylim(*property_info.plot_yrange())

        add_xy_labels(property_info, request)

        return finish(request)

@sessionfree_lru_cache(100)
def _get_property_from_halo_and_name(halo, name):
    return halo.calculate(name, True)

@view_config(route_name='array_csv', renderer='csv')
def array_csv(request):
    halo = halo_from_request(request)
    name = decode_property_name(request.matchdict['nameid'])
    val, property_info = halo.calculate(name, True)
    xval = property_info.plot_x_values(val)

    return {
        'header': ["bin_center", name],
        'rows': np.array((xval,val)).T,
    }



from .message import Message

_remote_sets = {}
class RemoteSetOperation(Message):
    def process(self):
        set_id, operation, value = self.contents
        global _remote_sets
        if operation=="add-if-not-exists":
            result = LocalSet(set_id).add_if_not_exists(value)
            self.reply(result)
        else:
            raise ValueError("Unknown operation %s" % operation)

    def reply(self, result):
        RemoteSetResult(result).send(self.source)

class RemoteSetResult(Message):
    pass

class SharedSet:
    def __new__(cls, set_id, allow_parallel=False):
        if cls is SharedSet:
            from . import parallelism_is_active
            parallel = allow_parallel and parallelism_is_active()
            if parallel:
                cls = RemoteSet
            else:
                cls = LocalSet
        return object.__new__(cls)

    def __init__(self, set_id):
        self.set_id = set_id

    def __getnewargs__(self):
        return self.set_id,

    def add_if_not_exists(self, value):
        """Adds the value to the set, and returns True if it was already present, as an atomic operation"""
        raise NotImplementedError("Constructing a SharedSet should automatically return a RemoteSet or LocalSet as appropriate")

class RemoteSet(SharedSet):
    def __init__(self, set_id, allow_parallel=False):
        assert allow_parallel
        super().__init__(set_id)
        self._underlying_set = set()

    def add_if_not_exists(self, value):
        """Adds to the set, and returns a boolean indicating whether the value was already present"""
        RemoteSetOperation((self.set_id, "add-if-not-exists", value)).send(0)
        result = RemoteSetResult.receive(0).contents
        return result

class LocalSet(SharedSet):
    def __init__(self, set_id, allow_parallel=False):
        super().__init__(set_id)
        self._underlying_set = _remote_sets.get(set_id, set())
        _remote_sets[set_id] = self._underlying_set
    def add_if_not_exists(self, value):
        result = value in self._underlying_set
        if not result:
            self._underlying_set.add(value)
        return result



import pynbody

from tangos import log
from tangos.parallel_tasks.message import Message
from tangos.util.check_deleted import check_deleted


class ConfirmLoadPynbodySnapshot(Message):
    pass


class PynbodySnapshotQueue:
    def __init__(self):
        self.timestep_queue = []
        self.handler_queue = []
        self.shared_mem_queue = []
        self.load_requester_queue = []
        self.current_timestep = None
        self.current_snapshot = None
        self.current_subsnap_cache = {}
        self.current_handler = None
        self.in_use_by = []

    def add(self, requester, handler, filename, shared_mem=False):
        log.logger.debug("Pynbody server: client %d requests access to %r", requester, filename)
        if shared_mem:
            log.logger.debug(" (shared memory mode)")
        if filename==self.current_timestep:
            self._notify_available(requester)
            self.in_use_by.append(requester)
        elif filename in self.timestep_queue:
            queue_position = self.timestep_queue.index(filename)
            self.load_requester_queue[queue_position].append(requester)
            assert self.handler_queue[queue_position] == handler
            assert self.shared_mem_queue[queue_position] == shared_mem
        else:
            self.timestep_queue.append(filename)
            self.handler_queue.append(handler)
            self.load_requester_queue.append([requester])
            self.shared_mem_queue.append(shared_mem)
        self._load_next_if_free()

    def free(self, requester):
        self.in_use_by.remove(requester)
        log.logger.debug("Pynbody server: client %d is now finished with %r", requester, self.current_timestep)
        self._free_if_unused()
        self._load_next_if_free()

    def get_subsnap(self, filter_or_object_spec, fam):
        if filter_or_object_spec is None:
            if fam is None:
                return self.current_snapshot
            else:
                return self.current_snapshot[fam]
        elif (filter_or_object_spec, fam) in self.current_subsnap_cache:
            log.logger.debug("Pynbody server: cache hit for %r (fam %r)",filter_or_object_spec, fam)
            return self.current_subsnap_cache[(filter_or_object_spec, fam)]
        else:
            log.logger.debug("Pynbody server: cache miss for %r (fam %r)",filter_or_object_spec, fam)
            subsnap = self.get_subsnap_uncached(filter_or_object_spec, fam)
            self.current_subsnap_cache[(filter_or_object_spec, fam)] = subsnap
            return subsnap

    def get_subsnap_uncached(self, filter_or_object_spec, fam):

        snap = self.current_snapshot

        if isinstance(filter_or_object_spec, pynbody.filt.Filter):
            snap = snap[filter_or_object_spec]
        elif isinstance(filter_or_object_spec, ObjectSpecification):
            snap = self.current_handler.load_object(self.current_timestep, filter_or_object_spec.object_number,
                                                    filter_or_object_spec.object_index,
                                                    filter_or_object_spec.object_typetag)
        else:
            raise TypeError("filter_or_object_spec must be either a pynbody filter or an ObjectRequestInformation object")

        if fam is not None:
            snap = snap[fam]

        return snap



    def _free_if_unused(self):
        if len(self.in_use_by)==0:
            from . import RequestPynbodyArray
            log.logger.info(
                f"Closing snapshot {self.current_timestep} after processing "
                f"{RequestPynbodyArray.get_num_requests()} array fetches")
            if RequestPynbodyArray.get_num_requests() > 0:
                log.logger.info("    Typical wait time to start retrieving an array: %.1fs +/- %.1fs",
                                RequestPynbodyArray.get_mean_wait_time(),
                                RequestPynbodyArray.get_std_wait_time())
                log.logger.info("    Total wait time: %.1fs", RequestPynbodyArray.get_total_wait_time())
                RequestPynbodyArray.reset_performance_stats()

            with check_deleted(self.current_snapshot):
                self.current_snapshot = None
                self.current_timestep = None
                self.current_subsnap_cache = {}
                self.current_handler = None

    def _notify_available(self, node):
        log.logger.debug("Pynbody server: notify %d that snapshot is now available", node)
        ConfirmLoadPynbodySnapshot(type(self.current_snapshot)).send(node)

    def _notify_unavailable(self, node):
        log.logger.debug("Pynbody server: notify %d that snapshot is unavailable", node)
        ConfirmLoadPynbodySnapshot(None).send(node)

    def _load_next_if_free(self):
        if len(self.timestep_queue)==0:
            return

        if self.current_handler is None:
            # TODO: Error handling
            self.current_timestep = self.timestep_queue.pop(0)
            self.current_handler = self.handler_queue.pop(0)
            self.current_shared_mem_flag = self.shared_mem_queue.pop(0)
            notify = self.load_requester_queue.pop(0)

            try:
                self.current_snapshot = self.current_handler.load_timestep(self.current_timestep)
                log.logger.info("Pynbody server: loaded %r", self.current_timestep)
                if self.current_shared_mem_flag:
                    log.logger.info("                (shared memory mode)")
                    self.current_snapshot._shared_arrays = True
                self.current_snapshot.physical_units()
                success = True
            except OSError:
                success = False

            if success:
                self.in_use_by = notify
                for n in notify:
                    self._notify_available(n)
            else:
                self.current_timestep = None
                self.current_handler = None
                self.current_snapshot = None

                for n in notify:
                    self._notify_unavailable(n)
                self._load_next_if_free()

        else:
            log.logger.info("The currently loaded snapshot is still required and so other clients will have to wait")
            log.logger.info("(Currently %d snapshots are in the queue to be loaded later)", len(self.timestep_queue))


_server_queue = PynbodySnapshotQueue()


class RequestLoadPynbodySnapshot(Message):
    def process(self):
        _server_queue.add(self.source, *self.contents)


class ReleasePynbodySnapshot(Message):
    def process(self):
        _server_queue.free(self.source)


_connection_active = False


class ObjectSpecification:
    def __init__(self, object_number, object_index, object_typetag='halo'):
        self.object_number = object_number
        self.object_index = object_index
        self.object_typetag = object_typetag

    def __repr__(self):
        return "ObjectSpecification(%d, %d, %r)"%(self.object_number, self.object_index, self.object_typetag)

    def __eq__(self, other):
        if not isinstance(other, ObjectSpecification):
            return False
        return self.object_number==other.object_number and self.object_typetag==other.object_typetag

    def __hash__(self):
        return hash((self.object_number, self.object_index, self.object_typetag))



from typing import List


def find_subclasses(cls: type) -> List[type]:
    subclasses = cls.__subclasses__()
    all_subclasses = []
    while subclasses:
        subclass = subclasses.pop()
        subclasses.extend(subclass.__subclasses__())
        all_subclasses.append(subclass)

    return all_subclasses



# Live properties suitable for calculations on underlying profiles, e.g. density profiles, mass profiles etc

import numpy as np

from . import LivePropertyCalculation


class AtPosition(LivePropertyCalculation):
    def __init__(self, simulation, position, array):
        super().__init__(simulation)
        self._array_info = array

    names = "at"

    def live_calculate(self, halo, pos, ar):
        return self._array_info.get_interpolated_value(pos, ar)



class MaxMinProperty(LivePropertyCalculation):
    def __init__(self, simulation, array):
        super().__init__(simulation)
        self._array_info = array

    names = "max", "min", "posmax", "posmin"

    def live_calculate(self, halo, array):
        max_, min_ = np.max(array), np.min(array)
        amax, amin = np.argmax(array), np.argmin(array)
        index_to_r = lambda index: index*self._array_info.plot_xdelta()+self._array_info.plot_x0()
        return float(max_), float(min_), index_to_r(amax), index_to_r(amin)



import gc
import pickle
import time

import numpy as np
import pynbody
import pynbody.snapshot.copy_on_access

import tangos.parallel_tasks.pynbody_server.snapshot_queue

from .. import log, remote_import
from ..message import ExceptionMessage, Message
from . import snapshot_queue, transfer_array
from .snapshot_queue import (ConfirmLoadPynbodySnapshot,
                             ReleasePynbodySnapshot,
                             RequestLoadPynbodySnapshot, _server_queue)


class ReturnPynbodyArray(Message):

    def __init__(self, contents, shared_mem = False):
        self.shared_mem = shared_mem
        super().__init__(contents)

    @classmethod
    def deserialize(cls, source, message):
        units, shared_mem = pickle.loads(message)

        contents = transfer_array.receive_array(source, use_shared_memory=shared_mem)

        if units is not None:
            if not isinstance(contents, pynbody.array.SimArray):
                contents = contents.view(pynbody.array.SimArray)
            contents.units = units

        obj = ReturnPynbodyArray(contents, shared_mem=shared_mem)
        obj.source = source

        return obj

    def serialize(self):
        assert isinstance(self.contents, np.ndarray)
        if hasattr(self.contents, 'units'):
            units = self.contents.units
        else:
            units = None

        serialized_info = pickle.dumps((units, self.shared_mem))
        return serialized_info

    def send(self, destination):
        # send envelope
        super().send(destination)

        # send contents
        transfer_array.send_array(self.contents, destination, use_shared_memory=self.shared_mem)

class RequestPynbodyArray(Message):
    _time_to_start_processing = []

    def __init__(self, filter_or_object_spec, array, fam=None, request_sent_time=None):
        self.filter_or_object_spec = filter_or_object_spec
        self.array = array
        self.fam = fam
        self.request_sent_time = request_sent_time

    @classmethod
    def deserialize(cls, source, message):
        obj = RequestPynbodyArray(*message)
        obj.source = source
        return obj

    def serialize(self):
        return (self.filter_or_object_spec, self.array, self.fam, time.time())

    @classmethod
    def get_mean_wait_time(cls):
        if len(cls._time_to_start_processing)==0:
            return 0
        else:
            return np.mean(cls._time_to_start_processing)

    @classmethod
    def get_total_wait_time(cls):
        if len(cls._time_to_start_processing)==0:
            return 0
        else:
            return np.sum(cls._time_to_start_processing)

    @classmethod
    def get_std_wait_time(cls):
        if len(cls._time_to_start_processing)==0:
            return 0
        else:
            return np.std(cls._time_to_start_processing)

    @classmethod
    def get_num_requests(cls):
        return len(cls._time_to_start_processing)

    @classmethod
    def reset_performance_stats(cls):
        cls._time_to_start_processing = []

    def process(self):
        start_time = time.time()
        self._time_to_start_processing.append(start_time - self.request_sent_time)

        try:
            log.logger.debug("Receive request for array %r from %d",self.array,self.source)
            subsnap = _server_queue.get_subsnap(self.filter_or_object_spec, self.fam)
            transfer_via_shared_mem = _server_queue.current_shared_mem_flag

            with subsnap.immediate_mode, subsnap.lazy_derive_off:
                if subsnap._array_name_implies_ND_slice(self.array):
                    raise KeyError("Not transferring a single slice %r of a ND array"%self.array)
                if self.array=='remote-index-list':
                    subarray = subsnap.get_index_list(subsnap.ancestor).view(pynbody.array.SimArray)
                    # this won't be actually in shared memory – it's a regular numpy array
                    transfer_via_shared_mem = False
                else:
                    subarray = subsnap[self.array]
                    assert isinstance(subarray, pynbody.array.SimArray)
                array_result = ReturnPynbodyArray(subarray, transfer_via_shared_mem)

        except Exception as e:
            array_result = ExceptionMessage(e)

        array_result.send(self.source)
        del array_result
        gc.collect()
        log.logger.debug("Array sent after %.2fs"%(time.time()-start_time))



class ReturnPynbodySubsnapInfo(Message):
    def __init__(self, families, sizes, properties, loadable_keys, fam_loadable_keys):
        super().__init__()
        self.families = families
        self.sizes = sizes
        self.properties = properties
        self.loadable_keys = loadable_keys
        self.fam_loadable_keys = fam_loadable_keys

    def serialize(self):
        return self.families, self.sizes, self.properties, self.loadable_keys, self.fam_loadable_keys

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(*message)
        obj.source = source
        return obj



class RequestPynbodySubsnapInfo(Message):
    def __init__(self, filename, filter_):
        super().__init__()
        self.filename = filename
        self.filter_or_object_spec = filter_

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(*message)
        obj.source = source
        return obj

    def serialize(self):
        return (self.filename, self.filter_or_object_spec)

    def process(self):
        start_time = time.time()
        assert(_server_queue.current_timestep == self.filename)
        if self.filter_or_object_spec is not None:
            log.logger.debug("Received request for subsnap info, spec %r", self.filter_or_object_spec)
        else:
            log.logger.debug("Received request for snapshot info")
        obj = _server_queue.get_subsnap(self.filter_or_object_spec, None)
        families = obj.families()
        fam_lengths = [len(obj[fam]) for fam in families]
        fam_lkeys = [obj.loadable_keys(fam) for fam in families]
        lkeys = obj.loadable_keys()
        ReturnPynbodySubsnapInfo(families, fam_lengths, obj.properties, lkeys, fam_lkeys).send(self.source)
        log.logger.debug("Info sent after %.2f",(time.time()-start_time))


class RemoteSnap(pynbody.snapshot.copy_on_access.UnderlyingClassMixin, pynbody.snapshot.SimSnap):
    def __init__(self, connection, filter_or_object_spec):
        """Create a remote snapshot object

        filter_or_object_spec can be:
        - a pynbody filter
        - a tuple (typetag, number) specifying an object to be loaded
        - None to load the whole snapshot (only sensible in shared memory mode)
        """
        super().__init__(connection.underlying_pynbody_class)
        self.connection = connection
        self._filename = connection.identity
        self._server_id = connection._server_id

        RequestPynbodySubsnapInfo(connection.filename, filter_or_object_spec).send(self._server_id)
        info = ReturnPynbodySubsnapInfo.receive(self._server_id)

        index = 0
        for fam, size in zip(info.families, info.sizes):
            self._family_slice[fam] = slice(index, index+size)
            index+=size
        self._num_particles = index

        self.properties.update(info.properties)
        self._loadable_keys = info.loadable_keys
        self._fam_loadable_keys = {fam: lk for fam, lk in zip(info.families, info.fam_loadable_keys)}
        self._filter_or_object_spec = filter_or_object_spec




    def _load_array(self, array_name, fam=None):
        RequestPynbodyArray(self._filter_or_object_spec, array_name, fam).send(self._server_id)
        try:
            start_time=time.time()
            log.logger.debug("Send array request")
            data = ReturnPynbodyArray.receive(self._server_id).contents
            log.logger.debug("Array received; waited %.2fs",time.time()-start_time)
        except KeyError:
            raise OSError("No such array %r available from the remote"%array_name)
        with self.auto_propagate_off:
            if len(data.shape)==1:
                ndim = 1
            elif len(data.shape)==2:
                ndim = data.shape[-1]
            else:
                assert False, "Don't know how to handle this data shape"

            if fam is None:
                self._create_array(array_name, ndim=ndim, source_array=data)
            else:
                self._create_family_array(array_name, fam, ndim=ndim, source_array=data)

    def _promote_family_array(self, name, *args, **kwargs):
        # special logic: the normal promotion procedure would copy everything for this array out of shared memory
        # which we don't want if the server can provide us with a shared memory view of the whole array

        if self.connection.shared_mem and not self.delay_promotion:
            if name in self._loadable_keys:
                for fam in self.families():
                    try:
                        del self[fam][name]
                    except KeyError:
                        pass
                self._load_array(name)
                return

        super()._promote_family_array(name, *args, **kwargs)






class RemoteSnapshotConnection:
    def __init__(self, input_handler, ts_extension, server_id=0, shared_mem=False):

        from ...input_handlers import pynbody
        assert isinstance(input_handler, pynbody.PynbodyInputHandler)

        if tangos.parallel_tasks.pynbody_server.snapshot_queue._connection_active:
            raise RuntimeError("Each client can only have one remote snapshot connection at any time")



        super().__init__()

        self._server_id = server_id
        self._input_handler = input_handler
        self.filename = ts_extension
        self.identity = "%d: %s"%(self._server_id, ts_extension)
        self.connected = False
        self.shared_mem = shared_mem

        # ensure server knows what our messages are about
        remote_import.ImportRequestMessage(__name__).send(self._server_id)

        log.logger.debug("Pynbody client: attempt to connect to remote snapshot %r", ts_extension)
        RequestLoadPynbodySnapshot((input_handler, ts_extension, self.shared_mem)).send(self._server_id)
        self.underlying_pynbody_class = ConfirmLoadPynbodySnapshot.receive(self._server_id).contents
        if self.underlying_pynbody_class is None:
            raise OSError("Could not load remote snapshot %r"%ts_extension)

        tangos.parallel_tasks.pynbody_server.snapshot_queue._connection_active = True
        self.connected = True
        log.logger.debug("Pynbody client: connected to remote snapshot %r", ts_extension)

        if self.shared_mem:
            self.shared_mem_view = self.get_view(None)

    def get_view(self, filter_or_object_spec):
        """Return a RemoteSubSnap that contains either the pynbody filtered region, or the specified object from a catalogue

        filter_or_object_spec is either an instance of pynbody.filt.Filter, or a tuple containing
        (typetag, number), which are respectively the object type tag and object number to be loaded
        """
        return RemoteSnap(self, filter_or_object_spec)

    def disconnect(self):

        if not self.connected:
            return

        ReleasePynbodySnapshot(self.filename).send(self._server_id)
        tangos.parallel_tasks.pynbody_server.snapshot_queue._connection_active = False
        self.connected = False

    def __del__(self):
        self.disconnect()



"""
Least-recently-used dictionary for caching

From https://gist.github.com/davesteele/44793cd0348f59f8fadd49d7799bd306

>>> import cache_dict
>>> c = cache_dict.CacheDict(cache_len=2)
>>> c[1] = 1
>>> c[2] = 2
>>> c[3] = 3
>>> c
CacheDict([(2, 2), (3, 3)])
>>> c[2]
2
>>> c[4] = 4
>>> c
CacheDict([(2, 2), (4, 4)])
>>>
"""

from collections import OrderedDict


class CacheDict(OrderedDict):
    """Dict with a limited length, ejecting LRUs as needed."""

    def __init__(self, *args, cache_len: int = 10, **kwargs):
        assert cache_len > 0
        self.cache_len = cache_len

        super().__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().move_to_end(key)

        while len(self) > self.cache_len:
            oldkey = next(iter(self))
            super().__delitem__(oldkey)

    def __getitem__(self, key):
        val = super().__getitem__(key)
        super().move_to_end(key)

        return val



import argparse

import numpy as np
from sqlalchemy import and_, or_, orm

from .. import core, query
from ..core import get_or_create_dictionary_item
from ..log import logger
from . import GenericTangosTool


class MergerTreePruner(GenericTangosTool):
    tool_name = 'prune-trees'
    tool_description = '[Experimental] Remove questionable links from merger trees'
    parallel = False

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')

        parser.add_argument('--min-weight-progenitor', action='store', type=float, default=0,
                            help='The minimum weight that should be associated with a progenitor for retention. '
                                 'To retain minor progenitors, this should normally be set to zero.')

        parser.add_argument('--min-weight-descendant', action='store', type=float, default=0.05,
                            help='The minimum weight that should be associated with a descendant for retention. '
                                 'Weights close to zero are normally suspect, since they indicate the halo'
                                 'transfers only a small fraction of its mass into the descendant.')

        parser.add_argument('--max-NDM-progenitor', action='store', type=float, default=2.0,
                            help='The maximum relative increase in number of dark matter particles in the progenitor'
                                 'for retention. For example, if this is set to 2.0 (default), the link is dropped'
                                 'if there are more than twice as many DM particles in the progenitor as in the'
                                 'descendant, since this suggests a mis-identification.')

        parser.add_argument('--min-NDM-progenitor', action='store', type=float, default=0.33,
                            help='The maximum relative  number of dark matter particles in the progenitor'
                                 'for retention. For example, if this is set to 0.33 (default), the link is dropped'
                                 'if there are less than a third as many DM particles in the progenitor as in the'
                                 'descendant, since this suggests a mis-identification.')


    def process_options(self, options: argparse.ArgumentParser):
        self.options = options

    def process_timestep(self, early_timestep: core.TimeStep):
        session = core.get_default_session()

        late_timestep = early_timestep.next
        if late_timestep is None:
            return


        descendant_alias = orm.aliased(core.halo.SimulationObjectBase, name="descendant")
        progenitor_alias = orm.aliased(core.halo.SimulationObjectBase, name="progenitor")
        progenitor_halolink = orm.aliased(core.HaloLink, name="descendant_to_progenitor_link")
        descendant_halolink = orm.aliased(core.HaloLink, name="progenitor_to_descendant_link")

        progenitor_links = session.query(progenitor_halolink.id, descendant_halolink.id)\
            .join(descendant_alias, and_(progenitor_halolink.halo_from_id == descendant_alias.id, descendant_alias.timestep_id == late_timestep.id))\
            .join(progenitor_alias, and_(progenitor_halolink.halo_to_id == progenitor_alias.id, progenitor_alias.timestep_id == early_timestep.id))\
            .outerjoin(descendant_halolink, and_(progenitor_halolink.halo_to_id == descendant_halolink.halo_from_id,
                                            progenitor_halolink.halo_from_id == descendant_halolink.halo_to_id))\
            .filter(or_(progenitor_halolink.weight < self.options.min_weight_progenitor,
                        descendant_halolink.weight < self.options.min_weight_descendant,
                        progenitor_alias.NDM > self.options.max_NDM_progenitor * descendant_alias.NDM,
                        progenitor_alias.NDM < self.options.min_NDM_progenitor * descendant_alias.NDM))


        delete_links = []
        progenitors_and_descendants = progenitor_links.all()
        for p,d in progenitors_and_descendants:
            delete_links.append(p)
            if d is not None:
                delete_links.append(d)

        row_count = session.query(core.HaloLink).filter(core.HaloLink.id.in_(delete_links)).delete()
        session.commit()
        if row_count>0:
            logger.info(f"Deleted {row_count} links between timesteps {early_timestep} and {late_timestep}")

    def run_calculation_loop(self):
        simulations = core.sim_query_from_name_list(self.options.sims)
        logger.info("This tool is experimental. Use at your own risk!")
        for simulation in simulations:
            logger.info("Processing %s", simulation)

            for ts in simulation.timesteps[::-1]:
                 self.process_timestep(ts)


class MergerTreePatcher(GenericTangosTool):
    tool_name = 'patch-trees'
    tool_description = "[Experimental] Attempt to patch up merger trees"
    parallel = False

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')
        parser.add_argument('--include-only', action='store', type=str, nargs='*', default=['NDM()>100'],
                            help="Specify a filter that describes which objects the calculation should be executed for. "
                                 "Multiple filters may be specified, in which case they must all evaluate to true for "
                                 "the object to be included.")
        parser.add_argument('--relative-match', action='store',  type=str, default=['NStar()','NDM()'],
                            help='Variables to minimise when finding a halo match. If multiple expressions'
                                 'are provided, their relative differences are added in quadrature')
        parser.add_argument('--max-for-match', action='store', type=float, default=0.1,
                            help='The maximum fractional offset in values of variables specified in relative-match'
                                 'that qualifies as a match')
        parser.add_argument('--position-variable', action='store', type=str, default='shrink_center/a()',
                            help='The name of the variable which represents the 3d position')
        parser.add_argument('--max-position-offset', action='store', type=float, default=100.0,
                            help='The maximum value of the position offset that is allowed in comoving units (i.e.'
                                 'will be divided by scalefactor)')
        parser.add_argument('--max-timesteps', action='store', type=int, default=6,
                            help='The maximum number of steps back in time to scan for a match')


    def process_options(self, options):
        self.options = options


    def create_link(self, halo, matched):
        session = core.get_default_session()
        core.creator.get_creator(session) # just to prevent any commits happening later when not expected

        d_id = get_or_create_dictionary_item(session, "patch-trees")
        logger.info(f"Creating a link from {halo} to {matched}")

        next = halo

        ts = halo.timestep.previous

        while ts!=matched.timestep:
            phantom = core.halo.PhantomHalo(ts, halo.halo_number, 0)
            # TODO: shouldn't assume this halo number is "free" for use, probably phantom halo IDs should be
            # allocated sequentially

            session.add(phantom)
            session.add(core.HaloLink(next, phantom, d_id, 1.0))
            session.add(core.HaloLink(phantom, next, d_id, 1.0))

            next = phantom
            ts = ts.previous

        session.add(core.HaloLink(next, matched, d_id, 1.0))
        session.add(core.HaloLink(matched, next, d_id, 1.0))
        session.commit()

    _candidates_cache = {}

    def _get_candidate_id_position_values(self, ts_candidate):
        if ts_candidate.id not in self._candidates_cache:
            self._candidates_cache[ts_candidate.id] = ts_candidate.calculate_all("dbid()",
                                                                      self.options.position_variable,
                                                                      *self.options.relative_match)
        return self._candidates_cache[ts_candidate.id]

    def fixup(self, halo):
        success = False
        ts_candidate = halo.timestep.previous
        if ts_candidate is None:
            return

        source_values = [halo.calculate(m) for m in self.options.relative_match]
        source_position = halo.calculate(self.options.position_variable)

        for i in range(self.options.max_timesteps):

            dbid, candidate_positions, *candidate_values = self._get_candidate_id_position_values(ts_candidate)

            offsets = [cv-sv for sv, cv in zip(source_values, candidate_values)]
            rel_offsets = [(o/sv)**2 for o, sv in zip(offsets, source_values)]
            rel_offsets = np.sqrt(sum(rel_offsets))

            pos_offset = np.linalg.norm(source_position - candidate_positions, axis=1)
            mask = pos_offset < self.options.max_position_offset*(1.+ts_candidate.redshift)

            dbid = dbid[mask]
            rel_offsets = rel_offsets[mask]

            if len(rel_offsets)>0:
                if np.min(rel_offsets) < self.options.max_for_match:
                    match_dbid = dbid[np.argmin(rel_offsets)]
                    match = query.get_halo(match_dbid)
                    self.create_link(halo,match)
                    success = True
                    break

            ts_candidate = ts_candidate.previous
            if ts_candidate is None:
                break

        if not success:
            logger.info(f"No luck in finding a match for {halo}")


    def run_calculation_loop(self):
        simulations = core.sim_query_from_name_list(self.options.sims)
        logger.info("This tool is experimental. Use at your own risk!")

        for simulation in simulations:
            logger.info("Processing %s",simulation)
            logger.info(self.options.include_only)
            include_criterion = "(!has_link(earlier(1))) & " + (" & ".join(self.options.include_only))

            logger.info("Query for missing links is %s",include_criterion)
            for ts in simulation.timesteps[::-1]:
                dbids, flag = ts.calculate_all("dbid()", include_criterion)
                dbids = dbids[flag]
                logger.info(f"Timestep {ts} has {len(dbids)} broken links to consider")
                for dbid in dbids:
                    obj = query.get_halo(dbid)
                    self.fixup(obj)



import sqlalchemy
from sqlalchemy import orm

from .. import core
from ..util import consistent_collection
from .multi_hop import MultiHopStrategy
from .one_hop import HopStrategy


class MultiSourceMultiHopStrategy(MultiHopStrategy):
    """A variant of MultiHopStrategy that finds halos corresponding to multiple start points.

    Note that the behaviour is necessarily somewhat different to the other classes which start from a single
    halo. Specifically, a target *must* be specified, and the direction of the hops to follow is inferred
    from the nature of the target.

    Additionally, as soon as any halo is "matched" in the target, the entire query is stopped. In other words,
    this class assumes that the number of hops is the same to reach all target halos."""

    def __init__(self, halos_from, target, **kwargs):
        """Construct a strategy for finding Halos via multiple "hops" along HaloLinks from multiple start-points

        :param halos_from: a list of all halos to start from.
        :param target: a TimeStep or Simulation object to target.
        :param one_match_per_input: if True (default), return one halo per starting point in order.
                                  The returned halo in each case should be the one with the
                                  highest weight link (i.e. the major progenitor or similar)

                                  if False, *all* linked halos are returned and the caller has to figure out
                                  which one belongs to which starting halo, e.g. by calling sources()

        Other parameters are passed onto an underlying MultiHopStrategy. However note that the order_by parameter
        has no effect unless one_match_per_input is False.
        """
        directed = kwargs.get("directed", self._infer_direction(halos_from, target))
        kwargs["directed"] = directed
        kwargs["target"] = target

        self._return_only_highest_weights = kwargs.pop('one_match_per_input', True)

        # For 'backwards' or 'forwards' searches (basically major progenitors or descendants), keep only the
        # strongest link at each _step_ rather than waiting to the end to select the highest weight.
        # This makes sure one never "hops" from one branch to another (see
        # test_hop_strategy.test_major_progenitor_from_minor_progenitor for an example that exposes
        # this former bug). The actual implementation of the per-step restriction is in the override to
        # _supplement_halolink_query_with_filter, below.
        self._keep_only_highest_weights_per_hop = (directed == "forwards" or directed == "backwards")
        self._keep_only_highest_weights_per_hop&=self._return_only_highest_weights | (target is None)

        super().__init__(halos_from[0], **kwargs)
        self._all_halo_from = halos_from

    def _infer_direction(self, halos_from, target):
        if isinstance(target, core.simulation.Simulation):
            return "across"
        elif isinstance(target, core.timestep.TimeStep):
            collected_halos = consistent_collection.ConsistentCollection(halos_from)
            if collected_halos.timestep.simulation_id!=target.simulation_id:
                return "across"
            elif collected_halos.timestep.time_gyr<target.time_gyr:
                return "forwards"
            else:
                return "backwards"


    def _seed_temp_table(self):
        insert_dictionaries = []
        for i,halo_from in enumerate(self._all_halo_from):
            insert_dictionaries.append({'halo_from_id': halo_from.id, 'halo_to_id': halo_from.id, 'weight': 1.0,
                                        'nhops': 0, 'source_id': i})

        self._connection.execute(self._table.insert(), insert_dictionaries)

    def _generate_next_level_prelim_links(self, from_nhops=0):
        if self._should_halt():
            return 0
        else:
            return super()._generate_next_level_prelim_links(from_nhops)

    def _supplement_halolink_query_with_filter(self, query, table=None):
        query = super()._supplement_halolink_query_with_filter(query,table)

        if self._keep_only_highest_weights_per_hop:
            query = self._extract_max_weight_rows_from_query(query, table)

        return query

    def _extract_max_weight_rows_from_query(self, query, table):
        from ..util.sql_argmax import argmax
        return argmax(query, table.c.weight, [table.c.source_id])


    def _should_halt(self):
        # should halt if there are some results available. We can tell this with the parent classes's
        # _generate_query. (Our own _generate_query always returns one result per source_id.)
        return super()._generate_query(True).count()>0

    def _order_by_clause(self, halo_alias, timestep_alias):
        if self._return_only_highest_weights:
            return [] # _return_only_highest weights is already ordered
        else:
            return [self._link_orm_class.source_id] \
               + super()._order_by_clause(halo_alias, timestep_alias)

    def all(self):
        results = self._get_query_all()
        if self._return_only_highest_weights:
            # query results include a source_id column which we now wish to ignore (see
            # _query_ordered below for more information)
            return [x[1].halo_to if x[1] is not None else None
                    for x in results]
        else:
            return [x.halo_to for x in results]

    def sources(self):
        """Returns the offset in the original list that generated each result returned by all().

        For example, if the class is constructed for two halos, but the results have two results for
        the first halo, sources() will return [0,0,1]."""
        results = self._get_query_all()
        if self._return_only_highest_weights:
            return [x[0] for x in results]
        else:
            return [x.source_id for x in results]

    def _generate_query(self, halo_ids_only):

        if self._return_only_highest_weights:
            # need to find the highest weight result for each source_id
            subquery = self._extract_max_weight_rows_from_query(super()._generate_query(halo_ids_only=False),
                                                                self._table).subquery()

            # we now need to restructure the results such that they appear in the same order as the input
            # halo list, and a NULL result is returned for 'missing' halos. This way, the result is
            # guaranteed to be in 1-1 correspondence with the input.
            #
            # We do this by joining to the initial seeds in the temp table

            subquery_orm_alias = orm.aliased(self._link_orm_class, subquery)
            source_ids = orm.aliased(self._link_orm_class)

            if halo_ids_only:
                # if going straight into a temptable, we only want to see the halo_to_id
                query = self.session.query(subquery_orm_alias.halo_to_id).select_from(source_ids)
            else:
                # if going back to SQLAlchemy ORM, we explicitly include the source_id, otherwise the sqlalchemy dedup
                # process removes duplicates rows (e.g. if there are several null results, only the
                # first will be returned!) resulting in a query result that is too short and unrecoverable
                # errors in functions that rely on getting back a 1-1 mapping
                query = self.session.query(source_ids.source_id,subquery_orm_alias).\
                    select_from(source_ids)

            query = query.filter(source_ids.nhops==0).order_by(source_ids.source_id)

            query = query.outerjoin(subquery_orm_alias,
                                    source_ids.source_id==subquery_orm_alias.source_id)

            if not halo_ids_only:
                query = query.options(orm.joinedload(subquery_orm_alias.halo_to))
        else:
            query = super()._generate_query(halo_ids_only)

        return query


class MultiSourceAllMajorProgenitorsStrategy(MultiSourceMultiHopStrategy):

    def __init__(self, halos_from, **kwargs):
        super().__init__(halos_from, None, one_match_per_input=False, directed='backwards',
                         include_startpoint=True)

    def _should_halt(self):
        return False

class MultiSourceAllMajorDescendantsStrategy(MultiSourceMultiHopStrategy):

    def __init__(self, halos_from, **kwargs):
        super().__init__(halos_from, None, one_match_per_input=False, directed='forwards',
                         include_startpoint=True)

    def _should_halt(self):
        return False



import contextlib
import random
import string
import sys

import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.orm
import sqlalchemy.orm.dynamic
import sqlalchemy.orm.query
from sqlalchemy import Column, Index, Integer, Table, and_
from sqlalchemy.orm import defer, relationship

from .. import config, core, temporary_halolist
from ..config import DOUBLE_PRECISION
from ..log import logger
from ..util.timing_monitor import TimingMonitor
from .one_hop import HopStrategy


class MultiHopStrategy(HopStrategy):
    """An extension of the HopStrategy class that takes multiple hops across
    HaloLinks, up to a specified maximum, before finding the target halo."""

    halo_old = sqlalchemy.orm.aliased(core.halo.SimulationObjectBase, name="halo_old")
    halo_new = sqlalchemy.orm.aliased(core.halo.SimulationObjectBase, name="halo_new")
    timestep_old = sqlalchemy.orm.aliased(core.timestep.TimeStep, name="timestep_old")
    timestep_new = sqlalchemy.orm.aliased(core.timestep.TimeStep, name="timestep_new")

    def __init__(self, halo_from, nhops_max=None, directed=None, target=None,
                 order_by=None, combine_routes=True, min_aggregated_weight=0.0,
                 min_onehop_weight=0.0, min_onehop_reverse_weight=None,
                 include_startpoint=False, one_simulation=None):
        """Construct a strategy for finding Halos via multiple "hops" along HaloLinks

        :param halo_from:   The halo to start hopping from
        :type halo_from:    core.SimulationObjectBase

        :param nhops_max:   The maximum number of hops to take

        :param directed:    The direction in which to step, which can be
              'backwards' - only take hop if it's backwards in time
              'forwards'  - only take hop if it's forwards in time
              'across'    - only take hop if it's to the same time
              None        - take all available hops

              Note that specifying a direction causes a slight slow-down in each hop because the hop results have
              to be joined to the TimeStep object to work out the direction. But the overall query times are
              often much shorter because the pruning cuts down the number of paths to be explored.

        :param target:      Only return results in target, which can be a TimeStep or Simulation object.
            If None, return all results. Intermediate hops are not restricted by target, only the results returned.

        :param one_simulation:  Only follow links to objects in the same simulation, if True. If False, follow
            links to any simulation. By default this is None, and the behaviour is dictated by directed: if
            directed is 'backwards' or 'forwards' one_simulation is True, whereas for other values it is False.

        :param order_by:    Return results in specified order; see HopStrategy for more information

        :param combine_routes:  If True (default), discard multiple routes to the same halo and keep only the
            strongest route. When False, one result is returned for each possible route,
            even if it ultimately reaches the same halo. This can result in a significant
            slow-down.

        :param min_aggregated_weight: Threshold cumulative weight below which routes will be discarded.
              The cumulative weight is defined as the product of the individual hop weights.
              Small values indicate that a small fraction of mass is in common.

        :param min_onehop_weight:     Threshold individual weight below which a route will be discarded.

        :param min_onehop_reverse_weight: Threshold individual weight for the link pointing in the opposite
                                          direction to that being followed, or None for no restriction.
                                          Note that this requires an extra join (if not None) and if there
                                          is no reverse link at all, the result will be dropped.

        :param include_startpoint:    Return the starting halo in the results (default False)
        """
        super().__init__(halo_from, target, order_by)
        if nhops_max is None:
            nhops_max = config.num_multihops_max_default
        self.nhops_max = nhops_max
        self.directed = directed
        self._min_aggregated_weight = min_aggregated_weight
        self._min_onehop_weight = min_onehop_weight
        self._min_onehop_reverse_weight = min_onehop_reverse_weight
        self._include_startpoint = include_startpoint
        if one_simulation is None:
            one_simulation = directed=='backwards' or directed=='forwards'
        self._one_simulation = one_simulation
        self._connection = self.session.connection()
        self._combine_routes = combine_routes
        self._debug_output = False # set to True to see information about discovered links as hops progress

        self.timing_monitor = TimingMonitor()
    def temp_table(self):
        """Execute the strategy and return results as a temp_table (see temporary_halolist module)"""
        if self._all is None:
            return self._temp_table_without_leaving_sql()
        else:
            return temporary_halolist.temporary_halolist_table(self.session, [x.halo_to_id for x in self._all])

    def _temp_table_without_leaving_sql(self):
        tt = self._manage_temp_table()
        tt.__enter__()
        try:
            self._generate_multihop_results()
        except:
            tt.__exit__(*sys.exc_info())
            raise

        thl = temporary_halolist.temporary_halolist_table(self.session,
                                                          self._order_query(self._generate_query(halo_ids_only=True)),
                                                          callback=lambda: tt.__exit__(None, None, None)
                                                          )
        return thl

    def _generate_multihop_results(self):
        self._seed_temp_table()
        self._make_hops()

    def _execute_query(self):
        with self._manage_temp_table():
            self._generate_multihop_results()
            try:
                q = self._order_query(self._generate_query(halo_ids_only=False))
                results = q.all()
                # NB the time here seems to be mainly making the ORM objects
                # rather than the query itself
            except sqlalchemy.exc.ResourceClosedError:
                results = []

        self._all = results


    def _supplement_halolink_query_with_filter(self, query, table=None):

        if table is None:
            table = core.halo_data.HaloLink.__table__

        if self._needs_join_for_link_filter():
            query = query. \
                join(self.halo_old, table.c.halo_from_id == self.halo_old.id). \
                join(self.halo_new, table.c.halo_to_id == self.halo_new.id). \
                join(self.timestep_old, self.halo_old.timestep). \
                join(self.timestep_new, self.halo_new.timestep)

        filter = self._generate_link_filter(self.timestep_old, self.timestep_new, table)
        query = query.filter(filter)

        return query

    def _needs_join_for_link_filter(self):
        return self.directed is not None

    def _generate_link_filter(self, timestep_old, timestep_new, table):

        recursion_filter = table.c.weight > self._min_aggregated_weight

        if self.directed is not None:
            directed = self.directed.lower()
            if self._one_simulation:
                recursion_filter &= (timestep_new.simulation_id == timestep_old.simulation_id)
            if directed == 'backwards':
                recursion_filter &= (timestep_new.time_gyr < timestep_old.time_gyr*(1.0-config.max_relative_time_difference))
            elif directed == 'forwards':
                recursion_filter &= (timestep_new.time_gyr > timestep_old.time_gyr*(1.0+config.max_relative_time_difference))
            elif directed == 'across':
                existing_timestep_ids = self.session.query(core.SimulationObjectBase.timestep_id).\
                    select_from(self._link_orm_class).join(self._link_orm_class.halo_to).distinct()
                recursion_filter &= ~timestep_new.id.in_(existing_timestep_ids)
                recursion_filter &= sqlalchemy.func.abs(timestep_new.time_gyr - timestep_old.time_gyr) \
                                    < config.max_relative_time_difference * timestep_old.time_gyr
            else:
                raise ValueError("Unknown direction %r" % directed)

        return recursion_filter

    def _delete_temp_table(self):
        self._table_index.drop(bind=self._connection)
        self._table.drop(checkfirst=True, bind=self._connection)
        self._prelim_table.drop(checkfirst=True, bind=self._connection)
        # self._index.drop(bind=self.connection)
        core.Base.metadata.remove(self._table)
        core.Base.metadata.remove(self._prelim_table)

    def _create_temp_table(self):
        rstr = ''.join(random.choice(string.ascii_lowercase) for _ in range(4))

        # The intent of the following two tables is that they are temporary. With SQLite, it is
        # essential that they are implemented literally as TEMPORARY tables as otherwise the performance
        # is hugely degraded. However, MySQL places two crippling limitations on TEMPORARY tables:
        #  1) No foreign keys allowed referring to columns in the permanent database
        #  2) Cannot join a TEMPORARY table to itself
        #
        # Restriction (1) can be evaded by not declaring the foreign key in the schema, and then
        # providing foreign_keys information in _construct_orm_class. However restriction (2) is
        # completely debilitating. Luckily the performance in MySQL is fine even if we don't
        # declare these tables as TEMPORARY, so we simply switch off the prefix.
        #
        # There is a strange third issue with MySQL which is that, if we declare the foreign key
        # when creating a table, the connection hangs. This seems to be because of a deadlock;
        # other open connections prevent creating the association to the existing tables (the
        # MySQL 'metadata lock'). So, even though MySQL will NOT be using temporary tables,
        # we still don't declare the foreign key dependence.

        dialect = self._connection.dialect.dialect_description.split("+")[0].lower()
        if dialect == 'mysql':
            prefixes = []
        else:
            prefixes = ['TEMPORARY']

        multi_hop_link_table = Table(
            'multihoplink_final_' + rstr,
            core.Base.metadata,
            Column('id', Integer, primary_key=True),
            Column('source_id', Integer),
            # Foreign keys below NOT declared at schema-level, see note above re MySQL
            Column('halo_from_id', Integer),
            Column('halo_to_id', Integer),
            Column('weight', DOUBLE_PRECISION),
            Column('nhops', Integer),
            prefixes = prefixes
        )

        multi_hop_link_prelim_table = Table(
            'multihoplink_prelim_' + rstr,
            core.Base.metadata,
            Column('id', Integer, primary_key=True),
            Column('source_id', Integer),
            # Foreign keys below NOT declared at schema-level, see note above re MySQL
            Column('halo_from_id', Integer),
            Column('halo_to_id', Integer),
            Column('weight', DOUBLE_PRECISION),
            Column('nhops', Integer),
            prefixes = prefixes
        )

        self._table_index = Index('temp.source_id_index_' + rstr, multi_hop_link_table.c.source_id, multi_hop_link_table.c.nhops)
        self._table_nhop_index = Index('temp.nhop_index_' + rstr, multi_hop_link_table.c.nhops)

        self._table = multi_hop_link_table
        self._prelim_table = multi_hop_link_prelim_table

        self._table.create(checkfirst=True, bind=self._connection)
        self._prelim_table.create(checkfirst=True, bind=self._connection)

    @contextlib.contextmanager
    def _manage_temp_table(self):
        self._create_temp_table()
        self._link_orm_class = self._construct_orm_class()
        yield
        self._delete_temp_table()

    def _seed_temp_table(self):
        insert_statement = self._table.insert().values(halo_from_id=self.halo_from.id, halo_to_id=self.halo_from.id,
                                    weight=1.0, nhops=0, source_id=0)

        self._connection.execute(insert_statement)

    def _make_hops(self):
        for i in range(0, self.nhops_max):
            with self.timing_monitor(self):
                self._nhops_taken = i
                generated_count = self._generate_next_level_prelim_links(i)
                if generated_count != 0:
                    filtered_count = self._filter_prelim_links_into_final()
                else:
                    filtered_count = 0

            # for performance info: self.timing_monitor.report_to_log(logger)

            if self._hopping_finished(filtered_count):
                break


    def _hopping_finished(self, filtered_count):
        return filtered_count==0

    def _generate_next_level_prelim_links(self, from_nhops=0):
        self.timing_monitor.mark('prelim-insert')
        new_weight = self._table.c.weight * core.halo_data.HaloLink.weight

        recursion_query = \
            self.session.query(
                core.halo_data.HaloLink.halo_from_id.label('halo_from_id'),
                core.halo_data.HaloLink.halo_to_id.label("halo_to_id"),
                new_weight.label('new_weight'),
                (self._table.c.nhops + sqlalchemy.literal(1)).label("nhops"),
                self._table.c.source_id.label('source_id')). \
                select_from(self._table). \
                outerjoin(core.halo_data.HaloLink, and_(self._table.c.nhops == from_nhops,
                                                           self._table.c.halo_to_id == core.halo_data.HaloLink.halo_from_id)). \
                filter(core.halo_data.HaloLink.weight > self._min_onehop_weight
                       )

        insert = self._prelim_table.insert().from_select(
            ['halo_from_id', 'halo_to_id', 'weight', 'nhops', 'source_id'],
            recursion_query)

        num_inserted = self._connection.execute(insert).rowcount

        self.timing_monitor.mark('prelim-thin')
        if self._combine_routes:
            # Ideally, before self._prelim_table.insert(), one would adapt recursion_query to return the argmax of
            # new_weight, grouped by halo_to_id and source_id. That could have been achieved using:
            #
            #   from ..util.sql_argmax import sql_argmax
            #   recursion_query = sql_argmax(recursion_query, "new_weight",
            #                               ["halo_to_id", "source_id"])
            #
            # However, this turns out to be very slow and there are no obvious ways to optimise it. It's essentially
            # using a huge CTE/subquery to replace a native argmax facility in SQL. Originally, with sqlite, we simply did:
            #
            #   recursion_query = recursion_query.group_by(core.halo_data.HaloLink.halo_to_id, self._table.c.source_id)
            #
            # It is actually unclear to me now why this ever worked, but anyway it doesn't work in MySQL. The balance of
            # remaining correct without killing performance is to actually insert all the rows into prelim_table, and
            # then delete the ones which are not wanted afterwards. This is presumably faster because the temp table
            # has relevant indices.
            from ..util.sql_argmax import delete_non_maximal_rows

            deleted_count = delete_non_maximal_rows(self._connection, self._prelim_table,
                                                    self._prelim_table.c.weight,
                                                    [self._prelim_table.c.halo_to_id, self._prelim_table.c.source_id])
            num_inserted-=deleted_count

        return num_inserted

    def _debug_print_links(self, tab):
        halo_from = sqlalchemy.orm.aliased(core.halo.Halo)
        halo_to = sqlalchemy.orm.aliased(core.halo.Halo)
        halo_source = sqlalchemy.orm.aliased(core.halo.Halo)
        q = self.session.query(halo_from, halo_to, tab.c.source_id, tab.c.weight, tab.c.nhops).\
            select_from(tab).\
            join(halo_from, tab.c.halo_from_id == halo_from.id).\
            join(halo_to, tab.c.halo_to_id == halo_to.id)
        for row in q.all():
            print(f"[s{row[2]} i{row[4]}] {row[0].path} -> {row[1].path} w={row[3]:.2f}")

    def _filter_prelim_links_into_final(self):
        self.timing_monitor.mark('final-insert')
        if self._debug_output:
            print()
            print(f"[{self._nhops_taken}] Preliminary links:")
            self._debug_print_links(self._prelim_table)

        q = self.session.query(self._prelim_table.c.halo_from_id,
                               self._prelim_table.c.halo_to_id,
                               self._prelim_table.c.weight,
                               self._prelim_table.c.nhops,
                               self._prelim_table.c.source_id)

        q = self._supplement_halolink_query_with_reverse_hop_filter(q, self._prelim_table)
        q = self._supplement_halolink_query_with_filter(q, self._prelim_table)

        added_rows = self._connection.execute(
            self._table.insert().from_select(['halo_from_id', 'halo_to_id', 'weight', 'nhops', 'source_id'],
                                             q)).rowcount

        if self._debug_output:
            logger.info(f"[{self._nhops_taken}] Accepted links:")
            self._debug_print_links(self._table)

        self.timing_monitor.mark('final-thin')
        self._connection.execute(self._prelim_table.delete())
        return added_rows

    def _supplement_halolink_query_with_reverse_hop_filter(self, query, table=None):
        if self._min_onehop_reverse_weight is None:
            return query

        if table is None:
            table = core.halo_data.HaloLink.__table__

        hl_alias = sqlalchemy.orm.aliased(core.HaloLink)

        query = query. \
                join(hl_alias,
                     and_(hl_alias.halo_from_id==table.c.halo_to_id,
                     hl_alias.halo_to_id==table.c.halo_from_id)).\
                filter(hl_alias.weight>self._min_onehop_reverse_weight)


        return query

    def _construct_orm_class(self):
        rstr = ''.join(random.choice(string.ascii_lowercase) for _ in range(4))
        class_name = "MultiHopHaloLink_"+rstr
        class_base = (core.Base,)
        class_attrs = {"__table__": self._table,
                       "halo_from": relationship(core.halo.SimulationObjectBase, primaryjoin=self._table.c.halo_from_id == core.halo.SimulationObjectBase.id,
                                                 foreign_keys = [self._table.c.halo_from_id]),
                       "halo_to"  : relationship(core.halo.SimulationObjectBase, primaryjoin=(self._table.c.halo_to_id == core.halo.SimulationObjectBase.id),
                                                 foreign_keys = [self._table.c.halo_to_id]),
                       "source_id" : self._table.c.source_id,
                       "weight" : self._table.c.weight,
                       "nhops" : self._table.c.nhops
                       }

        return type(class_name,class_base,class_attrs)


    def _generate_query(self, halo_ids_only):
        if halo_ids_only:
            query = self.session.query(self._link_orm_class.halo_to_id)
        else:
            query = self.session.query(self._link_orm_class)

        if not self._include_startpoint:
            query = query.filter(self._table.c.nhops>0)

        query = self._filter_query_for_target(query, self._target)

        return query

    def _generate_order_arg_from_name(self, name, halo_alias, timestep_alias):
        if name == 'nhops':
            return self._link_orm_class.c.nhops
        else:
            return super()._generate_order_arg_from_name(name, halo_alias, timestep_alias)



import importlib
import re
import sys
import warnings

from .. import config, core

backend = None
_backend_name = config.default_backend
_num_procs = None # only for multiprocessing backend

_on_exit = [] # list of functions to call when parallelism is shutting down

from .. import log
from . import accumulative_statistics, jobs, message


def use(name):
    global backend, _backend_name, _num_procs
    if backend is not None:
        warnings.warn("Attempt to specify backend but parallelism is already initialised. This call may have no effect, unless you know exactly what you're doing.", RuntimeWarning)
    if "-" in name:
        _backend_name, num_procs = name.split("-")
        _num_procs = int(num_procs)
    else:
        _backend_name = name

def _process_command_line():
    command_line = " ".join(sys.argv)
    match = re.match("^(.*)--backend *=? *([^ ]*)(.*)$", command_line)
    if match is not None:
        use(match.group(2))
        new_command_line = match.group(1)+" "+match.group(3)
        sys.argv = new_command_line.split()

_process_command_line()

def init_backend():
    global backend
    if backend is None:
        backend = importlib.import_module("."+_backend_name, package='tangos.parallel_tasks.backends')

def deinit_backend():
    global backend
    backend = None

def parallelism_is_active():
    global _backend_name
    return _backend_name != 'null' and backend is not None

def launch(function, args=None, backend_kwargs=None):
    if args is None:
        args = []

    if backend_kwargs is None:
        backend_kwargs = {}

    result = None

    # we need to close any existing connections because we may fork, which leads to
    # buggy/unreliable behaviour. This should invalidate the session attached to
    # any existing objects, which is intended behaviour. If you are using parallel
    # tasks, you need to re-query for any objects you are using once inside the
    # parallel jobs.
    if core._engine is not None:
        connection_info = core._internal_session_args
    else:
        connection_info = None

    try:
        init_backend()

        try:
            core.close_db()
            if _backend_name != 'null':
                result = backend.launch(_exec_function_or_server, [function, connection_info, args], **backend_kwargs)
            else:
                result = function(*args)
        finally:
            if connection_info is not None:
                core.init_db(*connection_info)
    finally:
        deinit_backend()

    return result

def distributed(file_list, proc=None, of=None):
    """Distribute a list of tasks between all nodes"""

    if type(file_list) == set:
        file_list = list(file_list)

    if _backend_name=='null':
        if proc is None:
            proc = 1
            of = 1
        i = (len(file_list) * (proc - 1)) // of
        j = (len(file_list) * proc) // of - 1
        assert proc <= of and proc > 0
        if proc == of:
            j += 1
        return file_list[i:j + 1]
    else:
        from . import jobs
        return jobs.parallel_iterate(file_list)


def _exec_function_or_server(function, connection_info, args):
    log.set_identity_string("[%3d] " % backend.rank())
    if connection_info is not None:
        log.logger.debug("Reinitialising database, args "+str(connection_info))
        core.init_db(*connection_info)

    if backend.rank()==0:
        _server_thread()
    else:
        function(*args)
        MessageExit().send(0)
    core.close_db()
    _shutdown_parallelism()
    log.set_identity_string("")

class MessageExit(message.Message):
    pass


def _server_thread():
    # Sit idle until request for a job comes in, then assign first
    # available job and move on. Jobs are labelled through the
    # provided iterator

    j = -1
    num_jobs = None
    current_job = None
    alive = [True for i in range(backend.size())]
    awaiting_barrier = [False for i in range(backend.size())]

    while any(alive[1:]):
        obj = message.Message.receive()
        if isinstance(obj, MessageExit):
            alive[obj.source]=False
        else:
            obj.process()

def on_exit_parallelism(function):
    global _on_exit
    _on_exit.append(function)

def _shutdown_parallelism():
    global backend, _on_exit
    log.logger.debug("Clearing up process")

    for fn in _on_exit:
        fn()
    _on_exit = []

    backend.barrier()
    backend.finalize()
    backend = None
    _bankend_name = 'null'




from . import remote_import, shared_set
from .barrier import barrier
from .lock import ExclusiveLock



from ...input_handlers import yt as yt_handler_module
from .. import PropertyCalculation


class YtPropertyCalculation(PropertyCalculation):
    works_with_handler = yt_handler_module.YtInputHandler
    requires_particle_data = True

from . import DM, basic, gas, stars



import re

from .. import config, core
from ..core import get_or_create_dictionary_item
from ..core.halo_data import HaloLink, HaloProperty
from ..input_handlers import pynbody
from ..log import logger
from ..util import timestep_object_cache
from . import GenericTangosTool


class SubfindTreeImporter(GenericTangosTool):
    tool_name = 'import-subfind-trees'
    tool_description = "Import merger trees calculated on-the-fly by Gadget/Arepo"
    parallel = False

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')


    def process_options(self, options):
        self.options = options


    def create_links(self, ts, ts_next):
        """Create the descendant and progenitor links between two timesteps"""
        session = core.get_default_session()
        d_id = get_or_create_dictionary_item(session, "subfind_tree_link")

        obj_cache = timestep_object_cache.TimestepObjectCache(ts)
        obj_cache_next = timestep_object_cache.TimestepObjectCache(ts_next)

        links = self._create_links(ts, obj_cache_next, d_id, "Desc")
        links += self._create_links(ts_next, obj_cache, d_id, "Prog")
        session.add_all(links)
        session.commit()
        logger.info("%d links created between %s and %s",len(links), ts, ts_next)

    def _create_links(self, starting_timestep, object_cache_other_timestep, dictionary_item, subfind_link_type):
        """Follow the SubFind linked list of descendants or progenitors"""
        links = []
        for high_z_halo in starting_timestep.halos.all():

            this_properties = high_z_halo.load(mode='subfind-properties')
            link_to_finder_id = this_properties['First'+subfind_link_type+'SubhaloNr']
            if link_to_finder_id<0:
                # I have found cases where FirstProgSubhaloNr = -1, but there *is* actually a progenitor, given by
                # ProgSubhaloNr. Equally, ProgSubhaloNr sometimes points to a progenitor which is not the main
                # progenitor, so we can't just ignore FirstProgSubhaloNr. I am really baffled by what the intended
                # purpose of ProgSubhaloNr is (as opposed to FirstProgSubhaloNr). However this is a pragmatic fix,
                # using it as a fallback where FirstProgSubhaloNr is borked (again, unsure why that happens).
                link_to_finder_id = this_properties[subfind_link_type+'SubhaloNr']
            this_mass = this_properties['SubhaloMass']
            while link_to_finder_id >= 0:
                this_descendant_tangos_obj = object_cache_other_timestep.resolve_from_finder_id(link_to_finder_id, 'halo')
                if this_descendant_tangos_obj is None:
                    # The subfind halo wasn't imported into tangos. The link therefore can't be imported.
                    # In principle, we could continue the search for next descendants with a bit of hacking,
                    # but in practice it is presumably pointless because we won't have *even smaller* things
                    # than this in the tangos db
                    break
                properties_of_linked_subhalo = this_descendant_tangos_obj.load(mode='subfind-properties')

                ratio = min(properties_of_linked_subhalo['SubhaloMass'] / this_mass, 1.0)

                links.append(HaloLink(high_z_halo, this_descendant_tangos_obj, dictionary_item, ratio))

                # now move onto next descendant
                link_to_finder_id = properties_of_linked_subhalo['Next'+subfind_link_type+'SubhaloNr']
        return links

    @classmethod
    def _get_snap_id(self, filename):
        return int(re.match(r".*_(\d\d\d)(.hdf5)?", filename).group(1))

    def run_calculation_loop(self):
        simulations = core.sim_query_from_name_list(self.options.sims)

        for simulation in simulations:
            assert isinstance(simulation.get_output_handler(), pynbody.GadgetSubfindInputHandler), \
                "This tool can only import trees from Gadget+Subfind simulations"

            logger.info("Processing %s",simulation)
            timesteps = simulation.timesteps
            for ts_prev, ts in zip(timesteps[:-1], timesteps[1:]):
                if self._get_snap_id(ts.extension)!=self._get_snap_id(ts_prev.extension)+1:
                    logger.warning("The snapshots appear to skip steps. Attempt to import tree will continue, but may result in errors or an inconsistent merger history.")
                self.create_links(ts_prev, ts)



from .multi_hop import MultiHopStrategy
from .multi_hop_variants import (MultiHopAllProgenitorsStrategy,
                                 MultiHopMajorDescendantsStrategy,
                                 MultiHopMajorProgenitorsStrategy,
                                 MultiHopMostRecentMergerStrategy)
from .multi_source import MultiSourceMultiHopStrategy
from .one_hop import (HopMajorDescendantStrategy, HopMajorProgenitorStrategy,
                      HopStrategy)



import os

import sqlalchemy
from sqlalchemy import Index, create_engine, event, inspect, text
from sqlalchemy.orm import clear_mappers, declarative_base, sessionmaker

from .. import config, log

_verbose = False
_internal_session=None
_internal_session_args=None
_internal_session_pid=None
_engine=None
Session=None


def get_default_session() -> sqlalchemy.orm.Session:
    """Get the default ORM session to be used when no other is specified."""
    global _internal_session, _internal_session_pid, _internal_session_args
    if _internal_session is None:
        init_db()
    elif _internal_session_pid!=os.getpid():
        if _internal_session_args is None:
            raise RuntimeError("The process has been forked, but no information is available to open a new connection to the database")
        else:
            init_db(*_internal_session_args)
    return _internal_session


def set_default_session(session):
    """Set the default ORM session to be used.

    No validation is carried out. If an inappropriate session is passed in, many database functions
    will subsequently fail. """
    global _internal_session
    _internal_session = session
    _internal_session_pid = os.getpid()
    _internal_session_args = None


def get_default_engine() -> sqlalchemy.engine.Engine:
    """Get the default sqlalchemy engine to be used when no other is specified."""
    global _engine
    if _engine is None:
        init_db()
    return _engine



clear_mappers()  # remove existing maps

Base = declarative_base()


from .creator import Creator
from .dictionary import DictionaryItem
from .halo import SimulationObjectBase
from .halo_data import HaloLink, HaloProperty
from .simulation import Simulation, SimulationProperty
from .timestep import TimeStep
from .tracking import TrackData, update_tracker_halos

Index("halo_index", HaloProperty.__table__.c.halo_id)
Index("name_halo_index", HaloProperty.__table__.c.name_id,
      HaloProperty.__table__.c.halo_id)
Index("halo_timestep_index", SimulationObjectBase.__table__.c.timestep_id)
Index("halo_creator_index", SimulationObjectBase.__table__.c.creator_id)
Index("halo_finder_index", SimulationObjectBase.__table__.c.finder_id)
Index("haloproperties_creator_index", HaloProperty.__table__.c.creator_id)
Index("halolink_index", HaloLink.__table__.c.halo_from_id)
Index("halolink_bidirectional_index", HaloLink.__table__.c.halo_to_id, HaloLink.__table__.c.halo_from_id)
Index("named_halolink_index", HaloLink.__table__.c.relation_id, HaloLink.__table__.c.halo_from_id)




def sim_query_from_name_list(names, session=None):
    if session is None:
        session = get_default_session()

    query = session.query(Simulation)
    clause = None
    if names is not None:
        for rxi in names:
            if clause is None:
                clause = (Simulation.basename == rxi)
            else:
                clause |= (Simulation.basename == rxi)

        query = query.filter(clause)

    return query

def sim_query_from_args(argv, session=None):
    if session is None:
        session = get_default_session()

    query = session.query(Simulation)

    if "--sims" in argv:
        rx = argv[argv.index("--sims") + 1:]
        clause = None
        for rxi in rx:
            if clause is None:
                clause = (Simulation.basename == rxi)
            else:
                clause = clause | (Simulation.basename == rxi)

        query = query.filter(clause)

    return query


def supplement_argparser(argparser):
    argparser.add_argument("--db-filename", "-d", help="Specify path to a database file to be used",
                           action='store', type=str, metavar="database_file.sqlite3")
    argparser.add_argument("--db-verbose", "-v", action='store_true',
                           help="Switch on sqlalchemy echo mode")


def process_options(argparser_options):
    global _verbose
    if argparser_options.db_filename is not None:
        config.db = argparser_options.db_filename
    _verbose = argparser_options.db_verbose

def _check_and_upgrade_database(engine, colname='finder_offset'):
    inspector = inspect(engine)
    if 'halos' in inspector.get_table_names():
        cols = inspector.get_columns('halos')
        for c in cols:
            if c['name']==colname:
                return
        log.logger.warning("The database uses an old schema, missing the finder_offset column from halos. Attempting to update.")
        with engine.begin() as connection:
            connection.execute(text(f"alter table halos add column {colname} integer;"))
            connection.execute(text(f"update halos set {colname} = finder_id;"))
        log.logger.warning("The database update appeared to complete without any problems.")


def init_db(db_uri=None, timeout=30, verbose=None):
    global _verbose, _internal_session, _engine, Session, _internal_session_args, _internal_session_pid
    if db_uri is None:
        db_uri = config.db

    if 'psycopg2' in db_uri:
        from ..util import postgresql_adapters
        postgresql_adapters.register_postgresql_adapters()

    if '//' not in db_uri:
        db_uri = 'sqlite:///' + db_uri

    if db_uri.startswith("sqlite:///"):
        connect_args = {"timeout": timeout}
    else:
        connect_args = {"connect_timeout": timeout}

    _engine = create_engine(
        db_uri,
        echo=verbose or _verbose,
        isolation_level='READ UNCOMMITTED',
        connect_args=connect_args,
        future=True
    )

    _check_and_upgrade_database(_engine)

    Session = sessionmaker(bind=_engine, future=True)
    _internal_session=Session()
    Base.metadata.create_all(_engine)
    creator.set_creator(None)
    _internal_session_args = (db_uri, timeout, verbose)
    _internal_session_pid = os.getpid() # stored so that we can detect when a fork happens

def close_db():
    global _engine
    close_session()
    if _engine is not None:
        _engine.dispose()
        _engine = None

def close_session():
    global Session, _internal_session
    if _internal_session is not None:
        _internal_session.close()
        _internal_session = None
    if Session is not None:
        Session = None

from .dictionary import (_get_dict_cache_for_session, get_dict_id,
                         get_or_create_dictionary_item)

__all__ = ['DictionaryItem',
           'sim_query_from_name_list', 'sim_query_from_args',
           'supplement_argparser',
           'update_tracker_halos',
           'process_options', 'init_db', 'Base', 'Creator',
           'get_default_session']



from .. import core


class SimulationGeneratorForTests:
    def __init__(self, sim_name="sim", session=None, max_steps=9):
        if not session:
            session = core.get_default_session()

        self.max_steps = max_steps
        self.session = session
        self.sim = core.simulation.Simulation(sim_name)
        self.session.add(self.sim)
        self._ptcls_in_common_dict = core.dictionary.get_or_create_dictionary_item(self.session, "ptcls_in_common")
        self._BH_dict = core.dictionary.get_or_create_dictionary_item(self.session, "BH")
        self._host_dict = core.dictionary.get_or_create_dictionary_item(self.session, "host")
        self.session.commit()

    def add_objects_to_timestep(self, num_halos, object_typecode=0, NDM=None):
        """Add halos to the most recently added timestep"""

        ts = self._most_recently_added_timestep()
        returned_halos = []

        for i in range(1,num_halos+1):
            if NDM is None:
                NDM_halo = 1000-i*100
            else:
                NDM_halo = NDM[i-1]
            halo = core.halo.SimulationObjectBase(ts, i, i, i, NDM_halo, 0, 0, 0)
            halo.object_typecode = object_typecode

            halo = core.halo.SimulationObjectBase(ts, i, i, i, NDM_halo, 0, 0, object_typecode=object_typecode)
            cl = core.halo.SimulationObjectBase.class_from_tag(
                core.halo.SimulationObjectBase.object_typetag_from_code(object_typecode)
            )
            halo.__class__ = cl # hmm, not really sure this should be necessary or advisable. But tests were failing
            # with the very specific combination of py3.11 + sqlalchemy 2.0.x + mysql, and this appears to fix it
            # so we better stick with it.
            returned_halos.append(halo)
        self.session.add_all(returned_halos)

        self.session.commit()
        return returned_halos

    def add_properties_to_halos(self, **properties):
        """Add properties to the halos in the most recently added timestep.

        For each kwarg, pass in a function mapping the halo number to the named property

        e.g. add_properties_to_halos(Mvir=lambda i: 100.*(10-i)) adds a Mvir value of 100*(10-halo_number)
         to each halo"""

        self._add_properties_to_objects(0, **properties)

    def _add_properties_to_objects(self, htype, **properties):
        """Add properties to the halos in the most recently added timestep.

        For each kwarg, pass in a function mapping the halo number to the named property

        e.g. add_properties_to_halos(Mvir=lambda i: 100.*(10-i)) adds a Mvir value of 100*(10-halo_number)
         to each halo"""

        for k, v in properties.items():
            for halo in self._most_recently_added_timestep().objects.filter_by(object_typecode=htype).all():
                halo[k] = v(halo.halo_number)

    def add_bhs_to_timestep(self, num_bhs):
        halos = self.add_objects_to_timestep(num_bhs, 1)
        return halos

    def add_properties_to_bhs(self, **properties):
        """Add properties to BHs. See add_properties_to_halos for documentation"""

        self._add_properties_to_objects(1, **properties)

    def _most_recently_added_timestep(self):
        return self.sim.timesteps[-1]

    def _two_most_recently_added_timesteps(self):
        return self.sim.timesteps[-2:]

    def add_timestep(self):
        """Add a sequentially-numbered timestep to the specified simulation"""

        timestep_num = len(self.sim.timesteps)+1
        if timestep_num>self.max_steps:
            raise ValueError("Number of steps added exceeds maximum. You can use max_steps=<n> when constructing SimulationGeneratorForTests.")
        ts = core.timestep.TimeStep(self.sim, "ts%d"%timestep_num)
        ts.redshift = self.max_steps - timestep_num
        ts.time_gyr = 0.9*timestep_num
        self.session.add(ts)
        self.session.commit()

        return ts

    def link_last_halos_using_mapping(self, mapping, adjust_masses=True, object_typecode=0) :
        """Store halolinks such that ts_source[halo_i].next = ts_dest[mapping[halo_i]]

        :type mapping dict"""

        ts_source, ts_dest = self._two_most_recently_added_timesteps()
        if adjust_masses:
            self._adjust_halo_NDM_for_mapping(mapping, ts_dest, ts_source)
        self._generate_bidirectional_halolinks_for_mapping(mapping, ts_dest, ts_source, object_typecode)
        self.session.commit()

    def link_last_halos(self, object_typecode=0, adjust_masses=False):
        """Generate default halolinks for the most recent two timesteps such that 1->1, 2->2 etc"""
        if len(self.sim.timesteps)<2:
            return
        ts_source, ts_dest = self._two_most_recently_added_timesteps()
        halo_nums_source = {a.halo_number for a in ts_source.objects.filter_by(object_typecode=object_typecode).all()}
        halo_nums_dest = {a.halo_number for a in ts_dest.objects.filter_by(object_typecode=object_typecode).all()}
        halo_nums_common = halo_nums_source.intersection(halo_nums_dest)
        mapping = {a:a for a in halo_nums_common}
        self.link_last_halos_using_mapping(mapping,adjust_masses, object_typecode)


    def link_last_bhs(self):
        self.link_last_halos(1)

    def link_last_bhs_using_mapping(self, mapping):
        self.link_last_halos_using_mapping(mapping, object_typecode=1)

    def assign_bhs_to_halos(self, mapping):
        ts = self._most_recently_added_timestep()

        for k,v in mapping.items():
            source_halo = ts.halos.filter_by(halo_number=v).first()
            target_bh = ts.bhs.filter_by(halo_number=k).first()

            forward_link = core.halo_data.HaloLink(source_halo, target_bh, self._BH_dict,
                                                   1.0)
            self.session.add(forward_link)

            reverse_link = core.halo_data.HaloLink(target_bh, source_halo, self._host_dict,
                                                   1.0)
            self.session.add(reverse_link)

        self.session.commit()

    def link_last_halos_across_using_mapping(self, other, mapping, object_typecode=0):
        ts_source = self._most_recently_added_timestep()
        ts_dest = other._most_recently_added_timestep()
        self._generate_bidirectional_halolinks_for_mapping(mapping, ts_dest, ts_source, object_typecode)
        self.session.commit()

    def add_mass_transfer(self, source_num, target_num, fraction_of_source, object_typecode=0, adjust_masses=False):
        ts_source, ts_dest = self._two_most_recently_added_timesteps()
        source_halo = ts_source.objects.filter_by(halo_number=source_num, object_typecode=object_typecode).first()
        target_halo = ts_dest.objects.filter_by(halo_number=target_num, object_typecode=object_typecode).first()

        transferred_ptcls = int(source_halo.NDM * fraction_of_source)
        fraction_of_dest = float(transferred_ptcls)/target_halo.NDM
        ts_source, ts_dest = self._two_most_recently_added_timesteps()

        if adjust_masses:
            target_halo.NDM+=source_halo.NDM * fraction_of_source
            major_descendant = source_halo.next
            if major_descendant:
                major_descendant.NDM-=source_halo.NDM * fraction_of_source

        forward_link = core.halo_data.HaloLink(source_halo, target_halo, self._ptcls_in_common_dict, fraction_of_source)
        backward_link = core.halo_data.HaloLink(target_halo, source_halo, self._ptcls_in_common_dict, fraction_of_dest)

        self.session.add_all([forward_link, backward_link])
        self.session.commit()



    def _generate_number_to_object_map(self, object_list):
        return {o.halo_number: o for o in object_list}
    def _generate_bidirectional_halolinks_for_mapping(self, mapping, ts_dest, ts_source, object_typecode):

        source_halos = self._generate_number_to_object_map(
            ts_source.objects.filter_by(object_typecode = object_typecode).all())
        target_halos = self._generate_number_to_object_map(
            ts_dest.objects.filter_by(object_typecode = object_typecode).all())

        links = []

        for source_num, target_num in mapping.items():
            source_halo = source_halos[source_num]
            target_halo = target_halos[target_num]

            forward_link = core.halo_data.HaloLink(source_halo, target_halo, self._ptcls_in_common_dict, 1.0)
            backward_link = core.halo_data.HaloLink(target_halo, source_halo, self._ptcls_in_common_dict, float(source_halo.NDM) / target_halo.NDM)

            links += [forward_link, backward_link]

        self.session.add_all(links)


    def _adjust_halo_NDM_for_mapping(self, mapping, ts_dest, ts_source):

        source_halos = self._generate_number_to_object_map(ts_source.halos.all())
        target_halos = self._generate_number_to_object_map(ts_dest.halos.all())

        for source_num, target_num in mapping.items():
            target_halo = target_halos[target_num]
            target_halo.NDM = 0

        for source_num, target_num in mapping.items():
            source_halo = source_halos[source_num]
            target_halo = target_halos[target_num]
            target_halo.NDM += source_halo.NDM



import abc
import argparse

from .. import core, parallel_tasks


class GenericTangosTool(metaclass=abc.ABCMeta):
    parallel = True
    tool_name = None
    tool_description = None

    def parse_command_line(self, argv=None):
        parser = self._get_parser_obj()
        options = parser.parse_args(argv)
        self.process_options(options)
        core.process_options(options)

    def _get_parser_obj(self):
        parser = self._create_parser_obj()
        self.add_parser_arguments(parser)
        return parser

    def _create_parser_obj(self):
        parser = argparse.ArgumentParser()
        core.supplement_argparser(parser)
        return parser

    @abc.abstractmethod
    def process_options(self, options):
        pass

    @abc.abstractmethod
    def run_calculation_loop(self):
        pass

    @classmethod
    @abc.abstractmethod
    def add_parser_arguments(cls, parser):
        pass

    @classmethod
    def _add_tool_to_subparser(cls, subparse):
        command = cls.tool_name
        help = cls.tool_description

        this_subparser = subparse.add_parser(command, help=help)
        cls.add_parser_arguments(this_subparser)

        def run(options):
            obj = cls()
            obj.process_options(options)
            if obj.parallel:
                parallel_tasks.launch(obj.run_calculation_loop,  [])
            else:
                obj.run_calculation_loop()

        this_subparser.set_defaults(func=run)

    @classmethod
    def add_tools(cls, subparse):
        if cls.tool_name is not None:
            cls._add_tool_to_subparser(subparse)
        for c in cls.__subclasses__():
            c.add_tools(subparse)

from . import (add_simulation, ahf_merger_tree_importer, changa_bh_importer,
               consistent_trees_importer, crosslink, db_importer,
               merger_tree_patcher, property_deleter, property_importer,
               property_writer, subfind_merger_tree_importer)



install_requires = [
    'setuptools',
    'numpy >= 1.10.0',
    'sqlalchemy >= 1.4.0', # moving towards sqlalchemy 2.0 eventually
    'pyparsing >= 2.1.0',
    'WebOb >= 1.7.0rc2', # Response.has_body
    'repoze.lru >= 0.4', # py3 compat
    'zope.interface >= 3.8.0',  # has zope.interface.registry
    'zope.deprecation >= 3.5.0', # py3 compat
    'venusian >= 1.0a3', # ``ignore``
    'translationstring >= 0.4', # py3 compat
    'PasteDeploy >= 1.5.0', # py3 compat
    'plaster',
    'plaster_pastedeploy',
    'pyramid',
    'waitress',
    'pyramid_debugtoolbar',
    'pyramid_jinja2',
    'pyramid_tm','pyramid_retry',
    'zope.sqlalchemy',
    'hupper',
    'scipy >= 0.14.0',
    'more_itertools >= 8.0.0',
    'matplotlib >= 3.0.0', # for web interface
    'tqdm >= 4.59.0',
    'tblib >= 3.0.0',
    ]

tests_require = [
    'pytest >= 5.0.0',
    'webtest >= 2.0',
    'pyquery >= 1.3.0',
    'pynbody >= 1.5.1',
    'yt>=3.4.0',
    'PyMySQL>=1.0.2',
    ]

import codecs
import os.path

from setuptools import find_packages, setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

setup(name='tangos',
      version=get_version("tangos/__init__.py"),
      description='TANGOS, the agile numerical galaxy organisation system',
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Intended Audience :: Developers",
          "Intended Audience :: Science/Research",
          "Topic :: Scientific/Engineering :: Astronomy",
          "Programming Language :: Python",
          "Programming Language :: Python :: 3.8",
          "Programming Language :: Python :: 3.9",
          "Programming Language :: Python :: 3.10",
          "Programming Language :: Python :: 3.11",
          "Programming Language :: Python :: 3.12",
          "License :: OSI Approved :: BSD License",
      ],
      author="Andrew Pontzen",
      author_email="a.pontzen@ucl.ac.uk",
      license="BSD",
      packages=find_packages(),
      entry_points={'paste.app_factory': [
                                            'main = tangos.web:main',
                                            ],
                    'console_scripts': [   'tangos_add_bh = tangos.scripts.add_bh:main',
                                           'tangos_bh_timelink = tangos.scripts.bh_timelink:main',
                                           'tangos_crosslink = tangos.scripts.crosslink:main',
                                           'tangos_fix_bh_hosts = tangos.scripts.fix_bh_hosts:main',
                                           'tangos_import_from_ahf = tangos.scripts.import_from_ahf:main',
                                           'tangos_manager = tangos.scripts.manager:main',
                                           'tangos_preprocess_bh = tangos.scripts.preprocess_bh:main',
                                           'tangos_timelink = tangos.scripts.timelink:main',
                                           'tangos_writer = tangos.scripts.writer:main',
                                           'tangos = tangos.scripts:main'
                    ]
      },
      include_package_data=True,
      zip_safe=False,
      python_requires='>=3.8',
      install_requires=install_requires,
      tests_require=tests_require,
      test_suite="tests",
      long_description=long_description,
      long_description_content_type='text/markdown',
      extras_require={'test': tests_require,
                      'rmdbs': ['PyMySQL[rsa]',
                                'psycopg2-binary']
                      }
      )



"""The proxy_object module provides a way to point to objects in the database without actually retrieving them.

This is used during property calculations to minimise the number of queries made to the database, and possibly to
point to objects that have not yet been created in the database at the time they are referred to."""

import abc

from .. import core


class ProxyResolutionException(Exception):
    """Unified exception raised when a proxy cannot be translated into an actual database object"""
    pass


class ProxyObjectBase(metaclass=abc.ABCMeta):
    """A proxy for an object (i.e. halo, group, BH etc) in the database"""

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def resolve(self, session):
        """Turn this ProxyObject into an actual object using the specified sqlalchemy session

        :type session: sqlalchemy.orm.Session
        """
        pass

    def relative_to_timestep_id(self, existing_timestep_id):
        """Return a proxy object resolved relative to an existing timestep, specified by database ID"""
        return self

    def relative_to_timestep_cache(self, timestep_cache):
        """Return a proxy object resolved using a cache of the timestep's objects.

        This results in more efficient database queries when a large number of objects are to be retrieved from the same
        timestep

        :type timestep_cache: tangos.util.timestep_object_cache.TimestepObjectCache"""
        return self

class ProxyObjectFromDatabaseId(ProxyObjectBase):
    """A proxy object that resolves into a database object with specified ID"""
    def __init__(self, dbid):
        super().__init__()
        self._dbid = dbid

    def resolve(self, session):
        return session.query(core.SimulationObjectBase).filter_by(id=self._dbid).first()

class ProxyObjectFromFinderIdAndTimestep(ProxyObjectBase):
    """A proxy object that resolves into the object with given finder ID in the specified timestep"""
    def __init__(self, finder_id, typetag, timestep_id):
        super().__init__()
        self._finder_id = finder_id
        self._typetag = typetag
        self._timestep_id = timestep_id

    def resolve(self, session):
        return session.query(core.SimulationObjectBase).filter_by(timestep_id=self._timestep_id, finder_id=self._finder_id,
                                                  object_typecode=core.SimulationObjectBase.object_typecode_from_tag(self._typetag)).first()

class ProxyObjectFromFinderIdAndTimestepCache(ProxyObjectBase):
    """A proxy object that resolves into the object with given finder ID in the specified timestep cache"""
    def __init__(self, finder_id, typetag, timestep_cache):
        super().__init__()
        self._finder_id = finder_id
        self._typetag = typetag
        self._timestep_cache = timestep_cache

    def resolve(self, session):
        if session!=self._timestep_cache.session:
            raise ProxyResolutionException("The session for the cache must match the session for the object resolution")
        return self._timestep_cache.resolve_from_finder_id(self._finder_id, self._typetag)

class IncompleteProxyObjectFromFinderId(ProxyObjectBase):
    """A proxy object that stores an object's finder ID and type, but requires the timestep still to be specified.

    By calling the method relative_to_timestep_id or relative_to_timestep_cache, a fully specified proxy object will
    be returned."""
    def __init__(self, finder_id, typetag):
        super().__init__()
        self._finder_id = finder_id
        self._typetag = typetag

    def resolve(self, session):
        raise ProxyResolutionException("Incomplete proxy object; the context must be specified")

    def relative_to_timestep_id(self, existing_timestep_id):
        return ProxyObjectFromFinderIdAndTimestep(self._finder_id, self._typetag, existing_timestep_id)

    def relative_to_timestep_cache(self, timestep_cache):
        return ProxyObjectFromFinderIdAndTimestepCache(self._finder_id, self._typetag, timestep_cache)



import zope.sqlalchemy

import tangos


def get_transaction_managed_session(session_factory, transaction_manager):
    """
    Get a ``sqlalchemy.orm.Session`` instance backed by a transaction.

    This function will hook the session to the transaction manager which
    will take care of committing any changes.

    - When using pyramid_tm it will automatically be committed or aborted
      depending on whether an exception is raised.

    - When using scripts you should wrap the session in a manager yourself.
      For example::

          import transaction

          engine = get_engine(settings)
          session_factory = get_session_factory(engine)
          with transaction.manager:
              dbsession = get_tm_session(session_factory, transaction.manager)

    """

    dbsession = session_factory()
    zope.sqlalchemy.register(
        dbsession, transaction_manager=transaction_manager)
    return dbsession


def includeme(config):
    """
    Initialize the model for a Pyramid app.

    Activate this setup using ``config.include('pyramid_scaffold.models')``.

    """
    settings = config.get_settings()
    settings['tm.manager_hook'] = 'pyramid_tm.explicit_manager'

    # use pyramid_tm to hook the transaction lifecycle to the request
    config.include('pyramid_tm')

    # use pyramid_retry to retry a request when transient exceptions occur
    config.include('pyramid_retry')

    config.registry['dbsession_factory'] = tangos.core.Session

    _ = tangos.core.get_default_engine()  # ensures the db is initialised

    # make request.dbsession available for use in Pyramid
    config.add_request_method(
        # r.tm is the transaction manager used by pyramid_tm
        lambda r: get_transaction_managed_session(tangos.core.Session, r.tm),
        'dbsession',
        reify=True
    )

    # add custom renderer for CSV files
    config.add_renderer('csv', 'tangos.web.renderers.CSVRenderer')



import os
import re

import tangos as db

from .. import config
from ..core import get_or_create_dictionary_item
from ..core.halo_data import HaloLink, HaloProperty
from ..input_handlers import ahf_trees as at
from ..log import logger
from . import GenericTangosTool


class AHFTreeImporter(GenericTangosTool):
    tool_name = 'import-ahf-trees'
    tool_description = "Import merger trees calculated by AHF's merger tree tools."
    parallel = False

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument('--sims', '--for', action='store', nargs='*',
                            metavar='simulation_name',
                            help='Specify a simulation (or multiple simulations) to run on')


    def process_options(self, options):
        self.options = options


    @staticmethod
    def filename_to_snapnum(filename):
        match = re.match(".*snapdir_([0-9]{3})/?", filename)
        if match:
            return int(match.group(1))
        else:
            raise ValueError("Unable to convert %s to snapshot number"%filename)

    def create_timestep_halo_dictionary(self, ts):
        halos = ts.halos.all()
        out = {}
        for h in halos:
            out[h.finder_id] = h
        return out

    def create_links(self, ts, ts_next, link_dictionary):
        session = db.get_default_session()
        d_id = get_or_create_dictionary_item(session, "ahf_tree_link")
        objs_this = self.create_timestep_halo_dictionary(ts)
        objs_next = self.create_timestep_halo_dictionary(ts_next)
        links = []
        for this_id, (next_id, merger_ratio) in link_dictionary:
            this_obj = objs_this.get(this_id, None)
            next_obj = objs_next.get(next_id, None)
            if this_obj is not None and next_obj is not None:
                links.append(HaloLink(this_obj, next_obj, d_id, merger_ratio))
                links.append(HaloLink(next_obj, this_obj, d_id, merger_ratio))
        session.add_all(links)
        session.commit()
        logger.info("%d links created between %s and %s",len(links), ts, ts_next)


    def run_calculation_loop(self):
        simulations = db.sim_query_from_name_list(self.options.sims)

        for simulation in simulations:
            logger.info("Processing %s",simulation)
            for ts in simulation.timesteps:
                ts_prev = ts.previous
                # ahf merger tree tool goes back in time
                if ts_prev is not None:
                    #additionally check if this is the first snapshot
                    tree = at.AHFTree(os.path.join(config.base,simulation.basename), ts)
                    self.create_links(ts_prev, ts, tree.get_links_for_snapshot())



from .. import log
from . import message

j = -1
num_jobs = None
current_job = None

class MessageStartIteration(message.Message):
    def process(self):
        global num_jobs, current_job
        if num_jobs is None:
            num_jobs = self.contents
            current_job = 0
        else:
            if num_jobs != self.contents:
                raise RuntimeError("Number of jobs (%d) expected by rank %d is inconsistent with %d" % (
                self.contents, self.source, num_jobs))


class MessageDeliverJob(message.Message):
    pass

class MessageDistributeJobList(message.Message):
    def process(self):
        # server should send this back out to all the other ranks
        from . import backend
        for rank in range(1, backend.size()):
            if rank != self.source:
                MessageDistributeJobList(self.contents).send(rank)
class MessageRequestJob(message.Message):
    def process(self):
        global j, num_jobs, current_job
        source = self.source
        if current_job is not None and num_jobs>0:
            log.logger.debug("Send job %d of %d to node %d", current_job, num_jobs, source)
        else:
            num_jobs = None
            current_job = None # in case num_jobs=0, still want to send 'end of loop' signal to client
            log.logger.debug("Finished jobs; notify node %d", source)

        MessageDeliverJob(current_job).send(source)

        if current_job is not None:
            current_job += 1
            if current_job == num_jobs:
                num_jobs = None
                current_job = None


def parallel_iterate(task_list):
    """Sets up an iterator returning items of task_list. """
    from . import backend, barrier

    assert backend is not None, "Parallelism is not initialised"
    MessageStartIteration(len(task_list)).send(0)
    barrier()

    while True:
        MessageRequestJob().send(0)
        job = MessageDeliverJob.receive(0).contents

        if job is None:
            barrier()
            return
        else:
            yield task_list[job]

def generate_task_list_and_parallel_iterate(task_list_function):
    """Call task_list_function on only one rank, and then parallel iterate with all ranks"""
    from . import backend

    assert backend is not None, "Parallelism is not initialised"

    if backend.rank()==1:
        task_list = task_list_function()
        MessageDistributeJobList(task_list).send(0)
        log.logger.debug("generated task list = %r",task_list)
    else:
        log.logger.debug("awaiting rank 1 generating task list")
        task_list = MessageDistributeJobList.receive(0).contents
        log.logger.debug("task_list = %r",task_list)
    return parallel_iterate(task_list)



import sys

import numpy as np

is_none = np.vectorize(lambda x: x is None, "?")
is_not_none = np.vectorize(lambda x: x is not None, "?")
is_false = np.vectorize(lambda x: x is False, "?")
is_not_false = np.vectorize(lambda x: x is not False, "?")



"""Routines for getting halo properties and links, and data derived from them, starting with a Halo or other object
"""

import sqlalchemy

from . import data_attribute_mapper


class HaloPropertyGetter:
    """HaloPropertyGetter and its subclasses implement efficient methods for retrieving data from sqlalchemy ORM objects.

    The key features are
     * being able to flexibly use a pre-queried cache of ORM objects with data, or issue fresh SQL queries as appropriate
     * being able to flexibly call 'reassembly' on the data
     * speed when used repeatedly on multiple rows of similar data

    Different classes get different types of data from the ORM and/or process it differently.

    This base class is used to retrieve the actual HaloProperty objects.
    """
    def use_fixed_cache(self, halo):
        return 'all_properties' not in sqlalchemy.inspect(halo).unloaded

    def get(self, halo, property_id, session):
        """Get the specified property, from the in-memory cache if it exists otherwise from the database
        using the specified session

        :type halo: SimulationObjectBase
        :type property_id: int
        :type session: sqlalchemy.orm.session.Session
        """
        if self.use_fixed_cache(halo):
            return self.get_from_cache(halo, property_id)
        else:
            return self.get_from_session(halo, property_id, session)

    def keys(self, halo, session):
        """Get a list of keys, from the in-memory cache if it exists otherwise from the database
            using the specified session

        :type halo: SimulationObjectBase
        :type session: sqlalchemy.orm.session.Session
        """
        if self.use_fixed_cache(halo):
            return self.keys_from_cache(halo)
        else:
            return self.keys_from_session(halo, session)

    def get_from_cache(self, halo, property_id):
        """Get the specified property from an existing in-memory cache

        :type halo: SimulationObjectBase
        :type property_id: int"""

        return_vals = []

        for x in halo.all_properties:
            if x.name_id == property_id:
                return_vals.append(x)

        return self.postprocess_data_objects(return_vals)


    def get_from_session(self, halo, property_id, session):
        """Get the specified property from the database using the specified session

        :type halo: SimulationObjectBase
        :type property_id: int
        :type session: sqlalchemy.orm.session.Session"""
        from . import halo_data
        query_properties = session.query(halo_data.HaloProperty).filter_by(name_id=property_id, halo_id=halo.id,
                                                                                        deprecated=False).order_by(
            halo_data.HaloProperty.id.desc())

        return self.postprocess_data_objects(query_properties.all())


    def keys_from_cache(self, halo):
        """Return a list of keys from an existing in-memory cache"""
        return [x.name.text for x in halo.all_properties]

    def keys_from_session(self, halo, session):
        from . import halo_data
        query_properties = session.query(halo_data.HaloProperty).filter_by(halo_id=halo.id,
                                                                                        deprecated=False)
        return [x.name.text for x in query_properties.all()]

    def cache_contains(self, halo, property_id):
        """Return True if the existing in-memory cache has the specified property

        :type halo: SimulationObjectBase
        :type property_id: int"""

        for x in halo.all_properties:
            if x.name_id == property_id:
                return True

        return False

    def postprocess_data_objects(self, objects):
        """Post-process the ORM data objects to pull out the data in the form required"""
        return objects



class HaloPropertyValueGetter(HaloPropertyGetter):
    """As HaloPropertyGetter, but return the data value (including automatic reassembly of the data if appropriate)"""
    def __init__(self):
        self._options = []
        self._providing_class = None
        self._mapper = None

    def postprocess_data_objects(self, outputs):
        return [self._postprocess_one_result(o) for o in outputs]

    def _setup_data_mapper(self, property_object):
        if self._mapper is None:
            # Optimisation: figure out a mapper for the first output and assume it's ok for all of them
            self._mapper = data_attribute_mapper.DataAttributeMapper(property_object)

    def _infer_property_class(self, property_object):
        if self._providing_class is None:
            # Optimisation: figure out a providing class for the first output and assume it's ok for all of them
            try:
                self._providing_class = property_object.name.providing_class(property_object.halo.handler_class)
            except NameError:
                pass

    def _postprocess_one_result(self, property_object):
        self._infer_property_class(property_object)

        if hasattr(self._providing_class, 'reassemble'):
            instance = self._providing_class(property_object.halo.timestep.simulation)
            return instance.reassemble(property_object, *self._options)
        else:
            self._setup_data_mapper(property_object)
            return self._mapper.get(property_object)



class HaloPropertyValueWithReassemblyOptionsGetter(HaloPropertyValueGetter):
    """As HaloPropertyValueGetter, but allow options to be passed to the property reassembler"""
    def __init__(self, *options):
        super().__init__()
        self._options = options

class HaloPropertyRawValueGetter(HaloPropertyValueGetter):
    """As HaloPropertyValueGetter, but never invoke an automatic reassembly; always retrieve the raw data"""
    def _postprocess_one_result(self, property_object):
        self._setup_data_mapper(property_object)
        return self._mapper.get(property_object)



class HaloLinkGetter(HaloPropertyGetter):
    """As HaloPropertyGetter, but retrieve HaloLinks instead of HaloProperties"""
    def get_from_cache(self, halo, property_id):
        return_vals = []
        for x in halo.all_links:
            if x.relation_id == property_id:
                return_vals.append(x)

        return self.postprocess_data_objects(return_vals)

    def get_from_session(self, halo, property_id, session):
        from . import halo_data
        query_links = session.query(halo_data.HaloLink).filter_by(relation_id=property_id, halo_from_id=halo.id).order_by(
            halo_data.HaloLink.id)
        return self.postprocess_data_objects(query_links.all())

    def cache_contains(self, halo, property_id):
        for x in halo.all_links:
            if x.relation_id == property_id:
                return True

        return False

    def keys_from_cache(self, halo):
        """Return a list of keys from an existing in-memory cache"""
        return [x.relation.text for x in halo.all_links]

    def keys_from_session(self, halo, session):
        from . import halo_data
        query_properties = session.query(halo_data.HaloLink).filter_by(halo_from_id=halo.id)
        return [x.relation.text for x in query_properties.all()]




class HaloLinkTargetGetter(HaloLinkGetter):
    """As HaloLinkGetter, but retrieve the target of the links instead of the HaloLink objects themselves"""
    def postprocess_data_objects(self, outputs):
        return [o.halo_to for o in outputs]



yt = None # deferred import; occurs when a YtInputHandler is constructed

import glob
import os

import numpy as np

from .. import config
from ..log import logger
from ..util.read_datasets_file import read_datasets
from . import HandlerBase, finding


class YtInputHandler(finding.PatternBasedFileDiscovery, HandlerBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        global yt
        import yt as yt_local
        yt = yt_local
        assert "3.4.0" <= yt.__version__, "Tangos requires yt version 3.4.0 or later"

    def get_timestep_properties(self, ts_extension):
        ts_filename =  self._extension_to_filename(ts_extension)
        f = yt.load(ts_filename)
        time_gyr = float(f.current_time.in_units("Gyr"))
        redshift = f.current_redshift
        results = {'time_gyr': time_gyr, 'redshift': redshift,
                   'available': True}
        return results

    def load_timestep_without_caching(self, ts_extension, mode=None):
        if mode!=None:
            raise ValueError("Custom load modes are not supported with yt")
        f = yt.load(self._extension_to_filename(ts_extension))
        return f

    def load_object(self, ts_extension, finder_id, finder_offset, object_typetag='halo', mode=None):
        f = self.load_timestep(ts_extension, mode)
        cat, cat_dat = self._load_halo_cat(ts_extension, object_typetag)
        center = cat_dat["halos","particle_position"][finder_offset]
        center+=f.domain_left_edge-cat.domain_left_edge
        radius = cat_dat["halos","virial_radius"][finder_offset]
        return f.sphere(center.in_cgs(), radius.in_cgs())


    def load_tracked_region(self, ts_extension, track_data, mode=None):
        raise NotImplementedError("Tracked regions not implemented for yt")

    def match_objects(self, ts1, ts2, halo_min, halo_max, dm_only=False, threshold=0.005, object_typetag='halo'):
        raise NotImplementedError("Matching halos still needs to be implemented for yt")

    def enumerate_objects(self, ts_extension, object_typetag="halo", min_halo_particles=config.min_halo_particles):
        if object_typetag!="halo":
            return
        if self._can_enumerate_objects_from_statfile(ts_extension, object_typetag):
            yield from self._enumerate_objects_from_statfile(ts_extension, object_typetag)
        else:
            logger.warning("No halo statistics file found for timestep %r", ts_extension)
            logger.warning(" => enumerating %ss directly using yt", object_typetag)

            catalogue, catalogue_data = self._load_halo_cat(ts_extension, object_typetag)
            num_objects = len(catalogue_data["halos", "virial_radius"])

            for i in range(num_objects):
                obj = self.load_object(ts_extension, i, i, object_typetag)
                NDM = len(obj["dark_matter","particle_mass"])
                NGas = len(obj["gas","mass"])
                NStar = len(obj["stars","particle_mass"])
                if NDM + NGas + NStar> min_halo_particles:
                    yield i, i, NDM, NStar, NGas


    def _load_halo_cat(self, ts_extension, object_typetag):
        snapshot_file = self.load_timestep(ts_extension)
        if object_typetag== 'halo':
            if not hasattr(snapshot_file, "_tangos_halos"):
                snapshot_file._tangos_halos = self._load_halo_cat_without_caching(ts_extension, snapshot_file)
            return snapshot_file._tangos_halos
        else:
            raise ValueError("Unknown halo type %r" % object_typetag)

    def _load_halo_cat_without_caching(self, ts_extension, snapshot_file):
        raise NotImplementedError("You need to select a subclass of YtInputHandler")

    def get_properties(self):
        return {}


class YtChangaAHFInputHandler(YtInputHandler):
    patterns = ["*.00???", "*.00????"]

    def _load_halo_cat_without_caching(self, ts_extension, snapshot_file):
        try:
            # yt 4
            from yt.frontends.ahf.api import AHFHalosDataset
        except ImportError:
            # yt 3
            from yt.frontends.ahf import AHFHalosDataset
        cat = AHFHalosDataset(self._extension_to_filename("halos/"+ts_extension)+".AHF_param",
                                                hubble_constant = snapshot_file.hubble_constant)
        cat_data = cat.all_data()
        return cat, cat_data


class YtEnzoRockstarInputHandler(YtInputHandler):
    patterns = ["RD????/RD????", "DD????/DD????"]
    auxiliary_file_patterns = ["halos_*.bin"]

    def load_timestep_without_caching(self, ts_extension, mode=None):
        from yt.data_objects.particle_filters import add_particle_filter
        if mode is not None:
            raise ValueError("Custom load modes are not supported with yt")
        f = yt.load(self._extension_to_filename(ts_extension))

        def Stars(pfilter, data):
            filter = data[("all", "particle_type")] == 2 # DM = 1, Stars = 2
            return filter

        add_particle_filter("stars", function=Stars, filtered_type='all', \
                            requires=["particle_type"])

        def AllDarkMatter(pfilter, data):
            filter = np.logical_or(data[("all", "particle_type")] == 4,data[("all", "particle_type")] == 1) # DM = 1, Stars = 2
            return filter

        add_particle_filter("dark_matter", function=AllDarkMatter, filtered_type='all', \
                            requires=["particle_type"])

        def MustRefineParticles(pfilter, data):
            filter = data[("all", "particle_type")] == 4
            return filter

        add_particle_filter("mrp_dark_matter", function=MustRefineParticles, filtered_type='all', \
                            requires=["particle_type"])

        f.add_particle_filter("stars")
        f.add_particle_filter("dark_matter")
        f.add_particle_filter("mrp_dark_matter")
        return f

    def _load_halo_cat_without_caching(self, ts_extension, snapshot_file):
        # Check whether datasets.txt exists (i.e., if rockstar was run with yt)
        if os.path.exists(self._extension_to_filename("datasets.txt")):
            fnum = read_datasets(self._extension_to_filename(""),ts_extension)
        else: # otherwise, assume a one-to-one correspondence
            overdir = self._extension_to_filename("")
            snapfiles = glob.glob(overdir+ts_extension[:2]+len(ts_extension[2:].split('/')[0])*'?')
            rockfiles = glob.glob(overdir+"out_*.list")
            sortind = np.array([int(rname.split('.')[0].split('_')[-1]) for rname in rockfiles])
            sortord = np.argsort(sortind)
            snapfiles.sort()
            rockfiles = np.array(rockfiles)[sortord]
            timestep_ind = np.argwhere(np.array([s.split('/')[-1] for s in snapfiles])==ts_extension.split('/')[0])[0]
            fnum = int(rockfiles[timestep_ind][0].split('.')[0].split('_')[-1])
        cat = yt.load(self._extension_to_filename("halos_"+str(fnum)+".0.bin"))
        cat_data = cat.all_data()
        # Check whether rockstar was run with Behroozi's distribution or Wise's
        if np.any(cat_data["halos","particle_identifier"]<0):
            del cat
            del cat_data
            cat = yt.load(self._extension_to_filename("halos_"+str(fnum)+".0.bin"))
            cat.parameters['format_revision'] = 2 #
            cat_data = cat.all_data()
        return cat, cat_data

    def enumerate_objects(self, ts_extension, object_typetag="halo", min_halo_particles=config.min_halo_particles):
        if object_typetag!="halo":
            return
        if self._can_enumerate_objects_from_statfile(ts_extension, object_typetag):
            yield from self._enumerate_objects_from_statfile(ts_extension, object_typetag)
        else:
            logger.warn("No halo statistics file found for timestep %r", ts_extension)
            logger.warn(" => enumerating %ss directly using yt", object_typetag)

            catalogue, catalogue_data = self._load_halo_cat(ts_extension, object_typetag)
            num_objects = len(catalogue_data["halos", "virial_radius"])

            for i in range(num_objects):
                obj = self.load_object(ts_extension, int(catalogue_data["halos","particle_identifier"][i]), i, object_typetag)
                NDM = len(obj["dark_matter","particle_mass"])
                NGas = 0 # cells
                NStar = len(obj["stars","particle_mass"])
                if NDM + NGas + NStar> min_halo_particles:
                    yield i, int(catalogue_data["halos","particle_identifier"][i]), NDM, NStar, NGas

    def load_object(self, ts_extension, finder_id, finder_offset, object_typetag='halo', mode=None):
        f = self.load_timestep(ts_extension, mode)
        cat, cat_dat = self._load_halo_cat(ts_extension, object_typetag)
        center = cat_dat["halos","particle_position"][cat_dat["halos","particle_identifier"]==finder_id][0]
        center+=f.domain_left_edge-cat.domain_left_edge
        radius = cat_dat["halos","virial_radius"][cat_dat["halos","particle_identifier"]==finder_id][0]
        return f.sphere(center.in_cgs(), radius.in_cgs())



import csv
from io import StringIO

# CSV renderer based on pyramid example
# https://docs.pylonsproject.org/projects/pyramid-cookbook/en/latest/templates/customrenderers.html

class CSVRenderer:
    def __init__(self, info):
        pass

    def __call__(self, value, system):
        """ Returns a plain CSV-encoded string with content-type
        ``text/csv``. The content-type may be overridden by
        setting ``request.response.content_type``."""
        request = system.get('request')
        if request is not None:
            response = request.response
            ct = response.content_type
            if ct == response.default_content_type:
                response.content_type = 'text/csv'
            if 'name' in value:
                response.content_disposition = 'attachment;filename=' + value['name'] + '.csv'

        fout = StringIO()
        writer = csv.writer(fout, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)

        writer.writerow(value.get('header', []))
        writer.writerows(value.get('rows', []))

        return fout.getvalue()



import contextlib
import inspect
import time

import numpy as np

from ..parallel_tasks import accumulative_statistics


class TimingMonitor(accumulative_statistics.StatisticsAccumulatorBase):
    """This class keeps track of how long a Property is taking to evaluate, and (if the Property is implemented
    to take advantage of this), the time spent on sub-tasks. It provides formatting to place this information
    into the log."""
    def __init__(self, allow_parallel=False):
        super().__init__(allow_parallel=allow_parallel)
        self.reset()
        self._monitoring = None

    @contextlib.contextmanager
    def __call__(self, object):
        self._start(object)
        yield
        self._end()

    def reset(self):
        self.timings_by_class = {}
        self.labels_by_class = {}

    def check_compatible_object(self, object):
        if not hasattr(object, 'timing_monitor'):
            raise TypeError("TimingMonitor requires a compatible object to monitor")

    def _set_as_monitor_for(self, object):
        self.check_compatible_object(object)
        if self._monitoring is not None:
            raise RuntimeError(
                f"TimingMonitor is already monitoring {self._monitoring!r} and cannot also monitor {object!r}")
        self._monitoring = object
        self._old_timing_monitor = object.timing_monitor
        object.timing_monitor = self

    def _unset_as_monitor_for(self, object):
        if self._monitoring is not object:
            raise RuntimeError("Consistency error - trying to stop monitoring an object that is not being monitored")
        object.timing_monitor = self._old_timing_monitor
        del self._old_timing_monitor
        self._monitoring = None

    def _start(self, object):
        """Start a timer for the specified object"""
        self._set_as_monitor_for(object)
        self._time_marks_info = ["start"]
        self._time_marks = [time.time()]

    def _end(self):
        """End a timer for the specified object."""
        cl = type(self._monitoring)
        self._unset_as_monitor_for(self._monitoring)
        self._time_marks_info.append("end")
        self._time_marks.append(time.time())

        self._add_run_to_running_totals(cl, self._time_marks, self._time_marks_info)

    def _add_run_to_running_totals(self, cl, latest_run_time_marks, latest_run_time_marks_labels):
        previous_timings = self.timings_by_class.get(cl, None)
        if previous_timings is None or len(previous_timings) == len(latest_run_time_marks) - 1:
            cumulative_timings = np.diff(latest_run_time_marks)
            if previous_timings is not None:
                cumulative_timings += previous_timings
            self.timings_by_class[cl] = cumulative_timings
            self.labels_by_class[cl] = latest_run_time_marks_labels
        else:
            # Incompatibility between this and previous timings from the same procedure. Can only track total time spent.
            start_time = latest_run_time_marks[0]
            end_time = latest_run_time_marks[-1]
            time_elapsed = end_time - start_time + sum(previous_timings)
            self.labels_by_class[cl] = ['start', 'end']
            self.timings_by_class[cl] = [time_elapsed]

    def mark(self, label=None):
        """Mark a named event so that more detailed timing can be given"""
        self._time_marks.append(time.time())
        if label is None:
            self._time_marks_info.append(
                inspect.currentframe().f_back.f_lineno)
        else:
            self._time_marks_info.append(label)

    def add(self, other):
        """Add the time taken by another TimingMonitor to this one"""
        if self._monitoring is not None:
            raise RuntimeError("Cannot add timings to a TimingMonitor that is currently monitoring a procedure")

        for c in other.labels_by_class.keys():
            labels = other.labels_by_class[c]
            timings = other.timings_by_class[c]
            self._add_run_to_running_totals(c, np.cumsum(np.concatenate(([0.0],timings))), labels)

    def report_to_log(self, logger):
        if len(self.timings_by_class) == 0:
            return
        logger.info("CUMULATIVE RUNNING TIMES")
        v_tot = 1e-10
        for k, v in self.timings_by_class.items():
            v_tot += sum(v)

        for k, v in self.timings_by_class.items():
            name = str(k)[:-2]
            name = name.split(".")[-1]
            name = "%20s " % (name[-20:])
            if len(v)>1:
                marks_info = self.labels_by_class[k]
                logger.info(" " + name + f"{sum(v):.1f}s | {100 * sum(v) / v_tot:.1f}%")
                logger.info("  ------ INTERNAL BREAKDOWN ------" )
                for i, this_v in enumerate(v):
                    logger.info((" %8s %8s %.1fs | %.1f%% | %.1f%%") %
                                (marks_info[i], marks_info[i + 1],
                                 this_v, 100 * this_v / sum(v), 100 * this_v / v_tot))
                logger.info("  --------------------------------")
            else:
                logger.info(name + f"{v[0]:.1f}s | {100 * v[0] / v_tot:.1f}%")



import copy
import logging
from io import StringIO

logger = logging.getLogger(__name__)

logger.setLevel(logging.INFO)
handler_stderr = logging.StreamHandler()
handler_stderr.setLevel(logging.INFO)
formatter = logging.Formatter("%(asctime)s : %(message)s")
handler_stderr.setFormatter(formatter)
logger.addHandler(handler_stderr)


class LogCapturer:
    def __init__(self):
        self.buffer = StringIO()
        self.handler_buffer = logging.StreamHandler(self.buffer)
        self.handler_buffer.setLevel(logging.INFO)
        self.handler_buffer.setFormatter(formatter)
        self._suspended_handlers = []

    def __enter__(self):
        self._suspended_handlers = copy.copy(logger.handlers)
        for x_handler in self._suspended_handlers:
            logger.removeHandler(x_handler)
        logger.addHandler(self.handler_buffer)
        return self

    def __exit__(self, *exc_info):
        for x_handler in self._suspended_handlers:
            logger.addHandler(x_handler)
        self._suspended_handlers = []
        logger.removeHandler(self.handler_buffer)

    def get_output(self):
        return self.buffer.getvalue()

    def get_output_without_timestamps(self):
        lines = self.get_output().split("\n")
        result = ""
        for l in lines:
            try:
                result += l.split(" : ", 1)[1]+"\n"
            except IndexError:
                result += l+"\n"
        return result


def set_identity_string(identifier):
    global handler_stderr
    formatter = logging.Formatter(identifier+"%(asctime)s : %(message)s")
    for handler in logger.handlers:
        handler.setFormatter(formatter)
